# encoding: utf-8
"""
@author: zyl
@file: my_utils.py
@time: ~~
@desc: zyl utils
"""

import re

import langid
import pandas as pd
from tqdm import tqdm


class MyTokenizer:
    def __init__(self):
        # 把连号‘-’分开
        self.sentence_tokenizer_zh = self._cut_paragraph_to_sentence_zh
        self.sentence_tokenizer_en = self._cut_paragraph_to_sentence_en().tokenize

        self.word_piece_tokenizer_zh = self._cut_sentence_to_word_piece_zh
        self.word_piece_tokenizer_en = self._cut_sentence_to_word_piece_en().tokenize

    def _cut_paragraph_to_sentence_zh(self, para: str, drop_empty_line=True, strip=True, deduplicate=False):
        """
        Args:
           para: 输入文本
           drop_empty_line: 是否丢弃空行
           strip:  是否对每一句话做一次strip
           deduplicate: 是否对连续标点去重，帮助对连续标点结尾的句子分句

        Returns:
           sentences: list of str
        """
        if deduplicate:
            para = re.sub(r"([。！？\!\?])\1+", r"\1", para)

        para = re.sub('([。！？\?!])([^”’])', r"\1\n\2", para)  # 单字符断句符
        para = re.sub('(\.{6})([^”’])', r"\1\n\2", para)  # 英文省略号
        para = re.sub('(\…{2})([^”’])', r"\1\n\2", para)  # 中文省略号
        para = re.sub('([。！？\?!][”’])([^，。！？\?])', r'\1\n\2', para)
        # 如果双引号前有终止符，那么双引号才是句子的终点，把分句符\n放到双引号后，注意前面的几句都小心保留了双引号
        para = para.rstrip()  # 段尾如果有多余的\n就去掉它
        # 很多规则中会考虑分号;，但是这里我把它忽略不计，破折号、英文双引号等同样忽略，需要的再做些简单调整即可。
        sentences = para.split("\n")
        if strip:
            sentences = [sent.strip() for sent in sentences]
        if drop_empty_line:
            sentences = [sent for sent in sentences if len(sent.strip()) > 0]
        return sentences

    def _cut_paragraph_to_sentence_en(self):
        from nltk.tokenize.punkt import PunktSentenceTokenizer, PunktParameters
        punkt_param = PunktParameters()
        abbreviation = ['et al.', 'i.e.', 'e.g.', 'etc.', 'i.e', 'e.g', 'etc', ' et al']
        punkt_param.abbrev_types = set(abbreviation)
        tokenizer = PunktSentenceTokenizer(punkt_param)
        return tokenizer

    def _cut_sentence_to_word_piece_zh(self, sentence: str):
        english = 'abcdefghijklmnopqrstuvwxyz0123456789αγβδεζηθικλμνξοπρστυφχψω'
        output = []
        buffer = ''
        for s in sentence:
            if s in english or s in english.upper():  # 英文或数字
                buffer += s
            else:  # 中文
                if buffer:
                    output.append(buffer)
                buffer = ''
                output.append(s)
        if buffer:
            output.append(buffer)
        return output

    def _cut_sentence_to_word_piece_en(self):
        from nltk import WordPunctTokenizer

        return WordPunctTokenizer()

    def cut_sentence_to_word_piece(self, sentence: str):
        if langid.classify(sentence)[0] == 'zh':
            return self.word_piece_tokenizer_zh(sentence)
        else:
            return self.word_piece_tokenizer_en(sentence)

    def cut_paragraph_to_sentences(self, paragraph: str):
        if langid.classify(paragraph)[0] == 'zh':
            return self.sentence_tokenizer_zh(paragraph)
        else:
            return self.sentence_tokenizer_en(paragraph)

class NlpUtils:
    def __init__(self):
        pass

    @staticmethod
    def show_all():
        import pandas as pd
        # 设置value的显示长度为200，默认为50
        pd.set_option('max_colwidth', 250)
        # 显示所有列，把行显示设置成最大
        pd.set_option('display.max_columns', None)
        # 显示所有行，把列显示设置成最大
        pd.set_option('display.max_rows', None)

    @staticmethod
    def df_clean_language(df, column_name, language_list=('en', 'zh')):
        # dataframe过滤出某一列文本的语言
        import langid
        df['language'] = df[column_name].apply(lambda x: langid.classify(str(x))[0])
        df = df[df['language'].isin(language_list)]
        df = df.drop(['language'], axis=1)
        return df

    @staticmethod
    def split_data_evenly(dt, num):
        dt_length = len(dt)
        step = int(dt_length / num)
        other_dt = dt_length % num

        if dt_length <= num:
            print('dt_length <= dt_num')
            return dt
        if other_dt == 0:
            return [dt[i:i + step] for i in range(0, dt_length, step)]
        else:
            first_dt = [dt[i:i + step + 1] for i in range(0, int((step + 1) * other_dt), step + 1)]
            second_list = [dt[i:i + step] for i in range(int((step + 1) * other_dt), dt_length, step)]
            first_dt.extend(second_list)
            return first_dt

    @staticmethod
    def clean_text(text):
        import re
        text = re.sub('<[^<]+?>', '', text).replace('\n', '').strip()  # 去html中的<>标签
        text = ' '.join(text.split()).strip()
        return text

    @staticmethod
    def cut_train_eval(all_df):
        from sklearn.utils import resample
        raw_df = resample(all_df, replace=False)
        cut_point = min(5000, int(0.2 * len(raw_df)))
        eval_df = raw_df[0:cut_point]
        train_df = raw_df[cut_point:]
        return train_df, eval_df

    @staticmethod
    def two_classification_sampling(train_df, column='labels', pos_label=1, mode='up_sampling'):
        import pandas as pd
        from sklearn.utils import resample
        negative_df = train_df[train_df[column] != pos_label]
        neg_len = negative_df.shape[0]
        positive_df = train_df[train_df[column] == pos_label]
        pos_len = positive_df.shape[0]
        if neg_len > pos_len:
            if mode == 'down_sampling':
                down_sampling_df = resample(negative_df, replace=False, n_samples=pos_len, random_state=242)
                return pd.concat([positive_df, down_sampling_df], ignore_index=True)
            else:

                up_sampling_df = resample(positive_df, replace=True, n_samples=(neg_len - pos_len), random_state=242)
                return pd.concat([train_df, up_sampling_df], ignore_index=True)
        elif neg_len < pos_len:
            if mode == 'down_sampling':
                down_sampling_df = resample(positive_df, replace=False, n_samples=neg_len, random_state=242)
                return pd.concat([down_sampling_df, negative_df], ignore_index=True)
            else:
                up_sampling_df = resample(negative_df, replace=True, n_samples=(pos_len - neg_len), random_state=242)
                return pd.concat([train_df, up_sampling_df], ignore_index=True)
        else:
            return train_df

    @staticmethod
    def find_index(raw_text, find_text, label='label'):
        # special_character = set(re.findall('\W', str(raw_text)))
        # for i in special_character:
        #     raw_text = raw_text.replace(i, '\\' + i)
        re_result = re.finditer(find_text, raw_text)
        starts = []
        for i in re_result:
            starts.append(i.span()[0])
        return [{'label': label, 'start': s, 'offset': len(find_text)} for s in starts]

    @staticmethod
    def ner_find(text: str, entities: dict, ignore_nested=True):
        """
        find the loaction of entities in a text
        Args:
            text: a text, like '我爱吃苹果、大苹果，小苹果，苹果【II】，梨子，中等梨子，雪梨，梨树。'
            entities: {'entity_type1':{entity_str1,entity_str2...},
                       'entity_type2':{entity_str1,entity_str2...},
                       ...}
                       like : {'apple': ['苹果', '苹果【II】'], 'pear': ['梨', '梨子'],}
            ignore_nested: if nested
        #>>>IndexedRuleNER().ner(text, entities, False)
        Returns:
            indexed_entities:{'entity_type1':[[start_index,end_index,entity_str],
                                              [start_index,end_index,entity_str]...]
                              'entity_type2':[[start_index,end_index,entity_str],
                                              [start_index,end_index,entity_str]...]
                                              ...}
        #>>>{'apple': [[3, 5, '苹果'], [7, 9, '苹果'], [11, 13, '苹果'], [14, 16, '苹果'], [14, 20, '苹果【II】']],
        'pear': [[21, 22, '梨'], [26, 27, '梨'], [30, 31, '梨'], [32, 33, '梨'], [21, 23, '梨子'], [26, 28, '梨子']]}
        """

        indexed_entities = dict()
        for every_type, every_value in entities.items():
            every_type_value = []
            for every_entity in list(every_value):
                special_character = set(re.findall('\W', str(every_entity)))
                for i in special_character:
                    every_entity = every_entity.replace(i, '\\' + i)
                re_result = re.finditer(every_entity, text)
                for i in re_result:
                    res = [i.span()[0], i.span()[1], i.group()]
                    if res != []:
                        every_type_value.append([i.span()[0], i.span()[1], i.group()])
            indexed_entities[every_type] = every_type_value
        if ignore_nested:
            for key, value in indexed_entities.items():
                all_indexs = [set(range(i[0], i[1])) for i in value]
                for i in range(len(all_indexs)):
                    for j in range(i, len(all_indexs)):
                        if i != j and all_indexs[j].issubset(all_indexs[i]):
                            value.remove(value[j])
                            indexed_entities[key] = value
                        elif i != j and all_indexs[i].issubset(all_indexs[j]):
                            value.remove(value[i])
                            indexed_entities[key] = value
        return indexed_entities

    @staticmethod
    def remove_some_model_files(args):
        import os
        if os.path.isdir(args.output_dir):
            cmd = 'rm -rf ' + args.output_dir.split('outputs')[0] + 'outputs/'
            os.system(cmd)
        if os.path.isdir(args.output_dir.split('outputs')[0] + '__pycache__/'):
            cmd = 'rm -rf ' + args.output_dir.split('outputs')[0] + '__pycache__/'
            os.system(cmd)
        if os.path.isdir(args.output_dir.split('outputs')[0] + 'cache/'):
            cmd = 'rm -rf ' + args.output_dir.split('outputs')[0] + 'cache/'
            os.system(cmd)

    # @staticmethod
    # def sunday_match(target, pattern):
    #     """
    #
    #     Args:
    #         target:
    #         pattern:
    #
    #     Returns:
    #
    #     """
    #     len_target = len(target)
    #     len_pattern = len(pattern)
    #
    #     if len_pattern > len_target:
    #         return list()
    #
    #     index = 0
    #     starts = []
    #     while index < len_target:
    #         if pattern == target[index:index + len_pattern]:
    #             starts.append(index)
    #             index += 1
    #         else:
    #             if (index + len(pattern)) >= len_target:
    #                 return starts
    #             else:
    #                 if target[index + len(pattern)] not in pattern:
    #                     index += (len_pattern + 1)
    #                 else:
    #                     index += 1
    #     return starts

    # @staticmethod
    # def transfomer_data_format_from_t5_to_ner(df: pd.DataFrame, delimiter='|',
    #                                           keep_addition_info=('id', 'text_type')):
    #     """
    #
    #     Args:
    #         df: dataframe,must have the columns-['prefix','input_text','target_text']
    #
    #     Returns:
    #
    #     """
    #     all_cls = df.value_counts('prefix').index.to_list()
    #     custom_labels = ['O']
    #     for c in all_cls:
    #         custom_labels.append('B-' + c.upper())
    #         custom_labels.append('I-' + c.upper())
    #     sentence_id = 0
    #     res_li = []
    #     my_tokenizer = MyTokenizer()
    #
    #     df = df.drop_duplicates(subset=['input_text'])
    #     for input_text, sub_df in tqdm(df.groupby('input_text', sort=False)):
    #         words = my_tokenizer.cut_sentence_to_word_piece(input_text)
    #         labels = ['O'] * len(words)
    #
    #         for _, d in sub_df.iterrows():
    #             if keep_addition_info:
    #                 for k in range(len(keep_addition_info)):
    #                     exec(f'info_{k} = d[keep_addition_info[{k}]]')
    #
    #             cls = d['prefix']
    #             sub_label = set(d['target_text'].split(delimiter))
    #             while '' in sub_label:
    #                 sub_label.remove('')
    #             if sub_label:
    #                 for every_entity in sub_label:
    #                     entity = my_tokenizer.cut_sentence_to_word_piece(every_entity)
    #                     res_starts = sunday_match(target=words, pattern=entity)
    #                     if res_starts:
    #                         for r in res_starts:
    #                             labels[r] = 'B-' + cls.upper()
    #                             if len(entity) > 1:
    #                                 labels[r + 1: r + len(entity)] = ['I-' + cls.upper()] * (len(entity) - 1)
    #
    #         sentence_ner = []
    #         for w, l in zip(words, labels):
    #             r = {'sentence_id': sentence_id, 'words': w, 'labels': l}
    #             if keep_addition_info:
    #                 for k in range(len(keep_addition_info)):
    #                     r.update({keep_addition_info[k]: eval(f'info_{k}')})
    #             sentence_ner.append(r)
    #
    #         res_li.extend(sentence_ner)
    #         sentence_id += 1
    #
    #     df = pd.DataFrame(res_li)
    #     return df


if __name__ == '__main__':
    test_df = pd.read_excel("/home/zyl/disk/PharmAI/pharm_ai/panel/data/v2.4.c/processed_0820.xlsx", 'eval')[0:100]

    print('1')
    # DTUtils.transfomer_data_format_from_t5_to_ner(test_df)
    pass
    # class Project(MyModel):
    #     def __init__(self):
    #         super(Project, self).__init__()
    #         self.start_time = '...'
    #         self.end_time = '...'
    #
    #         self.wandb_proj = 'test'
    #         self.use_model = 'classification'  # mt5 /classification
    #         self.model_type = 'bert'
    #         self.pretrained_model = ConfigFilePaths.bert_dir_remote
    #
    #     def run(self):
    #         self.train_test()
    #
    #     def train_test(self):
    #         self.model_version = 'vtest'
    #         self.pretrained_model = '/home/zyl/disk/PharmAI/pharm_ai/po/best_model/v4.2.0.4/'
    #         self.args = MyModel.set_model_parameter(model_version=self.model_version,
    #                                                 args=ClassificationArgs(),
    #                                                 save_dir='po')
    #         os.environ["CUDA_VISIBLE_DEVICES"] = "1,2,3"
    #         self.cuda_device = 0
    #         self.args.n_gpu = 3
    #
    #         self.args.num_train_epochs = 1
    #         self.args.learning_rate = 5e-5
    #         self.args.train_batch_size = 64  # 512
    #         self.args.eval_batch_size = 32  # 256
    #         self.args.max_seq_length = 512
    #         self.args.gradient_accumulation_steps = 8  # 256
    #
    #         train_df = pd.read_excel('./data/processed_0825.xlsx', 'train')
    #         eval_df = pd.read_excel('./data/processed_0825.xlsx', 'test')
    #         self.train(train_df=train_df, eval_df=eval_df)
    #
    #
    # pass
    # # d = range(0, 10)
    # # num = 5
    # # print(DTUtils.split_data_evenly(d, 5))
    # # print('1')
    # r = ['a',' ','','df','x',]
    # f = ['','df']
    # g = DTUtils.find_index(r, f)
    # print(g)
    # for i in g:
    #     print(r[i['start']:i['start']+i['offset']])
    # print(r[22:25])



# encoding: utf-8
'''
@author: zyl
@file: entry_match.py
@time: 2021/11/11 9:58
@desc:
'''

pass
# ##################################################################
# @staticmethod
# def eval_entry_match(model, eval_df: pd.DataFrame, my_dict, delimiter='|', use_dict_match=True,
#                      pos_neg_ratio=None, keep_entry_in_dict=True, use_multi_gpus=None):
#     prefixes = eval_df['prefix'].tolist()
#     input_texts = eval_df['input_text'].tolist()
#     target_texts = eval_df['target_text'].tolist()
#
#     revised_target_texts = NERUtils.em_revise_target_texts(prefixes=prefixes, target_texts=target_texts,
#                                                            prefix_dict=my_dict.prefix_dict,
#                                                            delimiter=delimiter,
#                                                            keep_entry_in_dict=keep_entry_in_dict)
#
#     pred_target_texts = NERUtils.predict_entry_match(em_model=model, prefix_match_dict=my_dict.prefix_match_dict,
#                                                      prefixes=prefixes, input_texts=input_texts,
#                                                      use_multi_gpus=use_multi_gpus,
#                                                      use_dict_match=use_dict_match)
#
#     revised_pred_target_texts = NERUtils.em_revise_target_texts(prefixes=prefixes, target_texts=pred_target_texts,
#                                                                 prefix_dict=my_dict.prefix_dict,
#                                                                 delimiter=delimiter,
#                                                                 keep_entry_in_dict=keep_entry_in_dict)
#
#     eval_df['true_target_text'] = revised_target_texts
#     eval_df['pred_target_text'] = revised_pred_target_texts
#
#     eval_res = {}
#     for prefix in set(prefixes):
#         prefix_df = eval_df[eval_df['prefix'] == prefix]
#         y_true = prefix_df['true_target_text'].tolist()
#         y_pred = prefix_df['pred_target_text'].tolist()
#         print(f'{prefix} report:')
#         res_df = NERUtils.entity_recognition_v2(y_true, y_pred, pos_neg_ratio=pos_neg_ratio)
#         eval_res[prefix] = res_df
#
#     print(f'sum report:')
#     res_df = NERUtils.entity_recognition_v2(revised_target_texts, revised_pred_target_texts,
#                                             pos_neg_ratio=pos_neg_ratio)
#     eval_res['sum'] = res_df
#     return eval_res
#
#
# @staticmethod
# def predict_entry_match(em_model, prefix_match_dict, prefixes: list, input_texts: list, use_dict_match=True,
#                         use_multi_gpus=None):
#     if len(input_texts) == 1:
#         use_multi_gpus = None
#     if use_dict_match:
#         pred_by_dict = []
#         for p, i in zip(prefixes, input_texts):
#             pred_by_dict.append(
#                 NERUtils.predict_entry_match_by_dict_match(str(i).strip(), dictionary=prefix_match_dict.get(p),
#                                                            use_edit_distance=False))
#
#             # i = i.lower()  # modify
#
#             # if p == 'disease_em':
#             #     pred_by_dict.append(
#             #         NERUtils.predict_entry_match_by_dict_match(i, dictionary=di_dict, use_edit_distance=False))
#             # else:
#             #     pred_by_dict.append(
#             #         NERUtils.predict_entry_match_by_dict_match(i, dictionary=tar_dict, use_edit_distance=False))
#     else:
#         pred_by_dict = [None] * len(input_texts)
#
#     to_predict_texts = [i + ': ' + j for i, j in zip(prefixes, input_texts)]
#     if not use_multi_gpus:
#         pred_by_model = em_model.predict(to_predict_texts)
#     else:
#         pred_by_model = em_model.predict_gpu(to_predict_texts, gpus=use_multi_gpus)
#     # pred_by_model = em_model.predict(to_predict_texts)
#     assert len(pred_by_model) == len(pred_by_dict)
#     pred_target_texts = [d if d else m for d, m in zip(pred_by_dict, pred_by_model)]
#     return pred_target_texts
#
#
# @staticmethod
# def predict_entry_match_by_dict_match(input_text: str, dictionary: dict, use_edit_distance: bool = False):
#     """predict the entry of a string by using dictionary match
#
#     Args:
#         input_text: a string
#         dictionary: the dict, {entity:entry}
#         use_edit_distance: True or False
#
#     Returns:
#         None or entry(str)
#     """
#     entry = dictionary.get(input_text)
#     if not entry:
#         if use_edit_distance:
#             import Levenshtein
#             max_score = 0
#             for every_entity in dictionary.keys():
#                 score = Levenshtein.ratio(every_entity, input_text)
#                 if score >= max_score and score > 0.80:  # 42-->43-->52
#                     max_score = score
#                 entry = dictionary.get(every_entity)
#     return entry  # None or entry
#
#
# @staticmethod
# def em_revise_target_texts(prefixes, target_texts, prefix_dict, delimiter='|', keep_entry_in_dict=False):
#     revised_target_texts = [NERUtils.revise_target_text(t_t, return_format='set', delimiter=delimiter) for
#                             t_t in target_texts]  # type:list[set,...]
#
#     if keep_entry_in_dict:
#         result = []
#         for p, r_t_t in zip(prefixes, revised_target_texts):
#             res = set()
#             if r_t_t:
#                 for j in list(r_t_t):
#                     if j in prefix_dict.get(p):
#                         res.add(j)
#             result.append(res)
#         return result
#     return revised_target_texts  # type:list[set]


    # @staticmethod
    # def eval_by_auto_batch_size(job, eval_df, initial_eval_batch_size=600):
    #     """
    #
    #     Args:
    #         job: you function. if run error, return None.
    #         eval_df: eval dataframe
    #         initial_eval_batch_size:
    #
    #     Returns:
    #
    #     """
    #     eval_batch_size = initial_eval_batch_size
    #     q = mp.Queue()
    #     pl = {'eval_batch_size': eval_batch_size}
    #     res = None
    #     while not res:
    #         eval_batch_size = int(eval_batch_size * 0.8)
    #         print(f'try eval_batch_size: {eval_batch_size}')
    #         pl['eval_batch_size'] = eval_batch_size
    #         eval_process = mp.Process(target=job, args=(pl, q, eval_df,))
    #         eval_process.start()
    #         eval_process.join()
    #         res = q.get()
    #         print(res)
    #
    # @staticmethod
    # def eval_by_different_parameters(job, parameter_cfg: dict, eval_df):
    #     q = mp.Queue()
    #     parameters_list = NERUtils.get_parameters_list(parameter_cfg)
    #     for pl in parameters_list:
    #         eval_process = mp.Process(target=job, args=(pl, q, eval_df,))
    #         eval_process.start()
    #         eval_process.join()
    #         print(q.get())
    #
    # @staticmethod
    # def get_parameters_list(parameter_cfg: dict):
    #     """
    #
    #     Args:
    #         parameter_cfg: like:{'truncating_size': [100,10], 'overlapping_size': [10],'max_seq_length':[100,30]}
    #
    #     Returns:[{'truncating_size': 100, 'overlapping_size': 10, 'max_seq_length': 100}, {'truncating_size': 100,
    #               'overlapping_size': 10, 'max_seq_length': 30}, {'truncating_size': 10, 'overlapping_size': 10,
    #               'max_seq_length': 100}, {'truncating_size': 10, 'overlapping_size': 10, 'max_seq_length': 30}]
    #
    #     """
    #     parameters_list = []
    #     keys = []
    #     values = []
    #     for i, j in parameter_cfg.items():
    #         keys.append(i)
    #         values.append(j)
    #     for para in product(*values):  # 求多个可迭代对象的笛卡尔积
    #         cfg = dict(zip(keys, para))
    #         parameters_list.append(cfg)
    #     return parameters_list  # type:list

    # @staticmethod
    # def cut_entities(input_entities: list, prefixes: list):
    #     assert len(input_entities) == len(prefixes)  # a input_text corresponds a prefix
    #     input_texts_ids = range(len(input_entities))
    #
    #     cut_ids = []
    #     cut_input_entities = []
    #     cut_prefixes = []

    #     for id, i_e, p in zip(input_texts_ids, input_entities, prefixes):
    #         if not isinstance(i_e, set):
    #             cut_i_e = NERUtils.revise_target_text(target_text=i_e, return_format='set', delimiter='|')
    #         else:
    #             cut_i_e = i_e
    #         if cut_i_e != set():
    #             for c_i_t in cut_i_e:
    #                 cut_ids.append(id)
    #                 cut_input_entities.append(c_i_t)
    #                 cut_prefixes.append(p)
    #     return cut_ids, cut_input_entities, cut_prefixes  # type:list
    #
    # @staticmethod
    # def combine_cut_entities(input_entities: list, cut_entities: list, cut_ids: list):
    #     dic = dict()
    #     for i, j in zip(cut_ids, cut_entities):
    #         if i not in dic.keys():
    #             dic[i] = j
    #         else:
    #             if isinstance(j, str):
    #                 dic[i] = dic[i] + '|' + j
    #             else:
    #                 dic[i].update(j)
    #
    #     res = []
    #     all_keys = list(dic.keys())
    #     for i in range(len(input_entities)):
    #         if i in all_keys:
    #             res.append(dic[i])
    #         else:
    #             res.append(set())
    #     return res
###################################
# eval_entry_match
#       em_revise_target_texts
#       predict_entry_match
#               predict_entry_match_by_dict_match
#               model.predict_gpu
#


# encoding: utf-8
'''
@author: zyl
@file: train_example.py
@time: 2021/11/11 16:14
@desc:
'''

import os

import pandas as pd

from zyl_utils.model_utils.models.my_model import MyModel


class Test(MyModel):
    def __init__(self):
        super(Test, self).__init__()
        self.start_time = '...'
        self.end_time = '...'

        self.wandb_proj = 'ttt'
        self.model_version = 'v0.0.0.0'  # to save model or best model

        self.use_model = 't5'  # t5/ mt5 /classification
        self.model_type = 't5'
        self.pretrained_model = '/large_files/pretrained_pytorch/t5-base/'  # 预训练模型位置

        self.use_cuda = True
        self.cuda_device = 0

        self.args = MyModel.set_model_parameter(model_version=self.model_version,
                                                args=self._set_args(),
                                                save_dir="/home/zyl/disk/test/")

    def run(self):
        self.train_test()

    def train_test(self):
        self.pretrained_model = '/home/zyl/disk/best_model/v4.2.0.4/'
        self.model_version = 'vtest'
        self.args = MyModel.set_model_parameter(model_version=self.model_version,
                                                args=self._set_args(), save_dir="/home/zyl/disk/test/")

        os.environ["CUDA_VISIBLE_DEVICES"] = "0,1"
        self.cuda_device = 0
        self.args.n_gpu = 2

        self.args.num_train_epochs = 1
        self.args.learning_rate = 5e-5
        self.args.train_batch_size = 16  # 512
        self.args.eval_batch_size = 32  # 256
        self.args.max_seq_length = 128
        self.args.gradient_accumulation_steps = 2  # 256

        train_df = pd.read_excel('./data/v1/train_1110.xlsx', 'train_mt5')[0:500]

        train_df.rename(columns={"therapy_labels_num": "target_text"}, inplace=True)
        train_df['prefix'] = 'classification'

        eval_df = pd.read_excel('./data/v1/train_1110.xlsx', 'eval_mt5')[0:200]

        eval_df.rename(columns={"therapy_labels_num": "target_text"}, inplace=True)
        eval_df['prefix'] = 'classification'

        self.train(train_df, eval_df)


if __name__ == '__main__':
    Test().run()



# encoding: utf-8
'''
@author: zyl
@file: my_DDPT5model.py
@time: 2021/11/11 11:00
@desc:
'''
import logging
import math
import os
import random
from dataclasses import asdict

import pandas as pd
import torch
import torch.multiprocessing as mp
import torch.nn.functional as F
from simpletransformers.t5.t5_model import T5Model
from tensorboardX import SummaryWriter
from torch.utils.data import DataLoader
from torch.utils.data.distributed import DistributedSampler
from tqdm.auto import tqdm, trange
from transformers.optimization import AdamW, Adafactor
from transformers.optimization import (
    get_constant_schedule,
    get_constant_schedule_with_warmup,
    get_linear_schedule_with_warmup,
    get_cosine_schedule_with_warmup,
    get_cosine_with_hard_restarts_schedule_with_warmup,
    get_polynomial_decay_schedule_with_warmup,
)

try:
    import wandb

    wandb_available = True
except ImportError:
    wandb_available = False

logger = logging.getLogger(__name__)


class DDPT5Model(T5Model):
    """The DDP version of T5Model"""

    def __init__(
            self,
            model_type,
            model_name,
            args=None,
            tokenizer=None,
            use_cuda=True,
            cuda_device=-1,
            **kwargs,
    ):

        """
        Initializes a DDP T5Model model. Turn off multi-processing settings.

        Args:
            model_type: The type of model (t5, mt5)
            model_name: The exact architecture and trained weights to use. This may be a Hugging Face Transformers compatible pre-trained model, a community model, or the path to a directory containing model files.
            args (optional): Default args will be used if this parameter is not provided. If provided, it should be a dict containing the args that should be changed in the default args.
            use_cuda (optional): Use GPU if available. Setting to False will force model to use CPU only.
            cuda_device (optional): Specific GPU that should be used. Will use the first available GPU by default.
            **kwargs (optional): For providing proxies, force_download, resume_download, cache_dir and other options specific to the 'from_pretrained' implementation where this will be supplied.
        """  # noqa: ignore flake8"
        super().__init__(model_type, model_name, args, tokenizer, use_cuda, cuda_device, **kwargs)
        self.args.use_multiprocessing = False
        self.args.use_multiprocessing_for_evaluation = False
        if self.args.n_gpu == 1:
            raise ValueError("You are using DDP with single GPU.")

    def train_model(
            self,
            train_data,
            output_dir=None,
            show_running_loss=True,
            args=None,
            eval_data=None,
            verbose=True,
            **kwargs,
    ):
        """
        Trains the model using 'train_data'

        Args:
            train_data: Pandas DataFrame containing the 3 columns - `prefix`, `input_text`, `target_text`.
                        - `prefix`: A string indicating the task to perform. (E.g. `"question"`, `"stsb"`)
                        - `input_text`: The input text sequence. `prefix` is automatically prepended to form the full input. (<prefix>: <input_text>)
                        - `target_text`: The target sequence
            output_dir: The directory where model files will be saved. If not given, self.args.output_dir will be used.
            show_running_loss (optional): Set to False to prevent running loss from being printed to console. Defaults to True.
            args (optional): Optional changes to the args dict of the model. Any changes made will persist for the model.
            eval_data (optional): A DataFrame against which evaluation will be performed when evaluate_during_training is enabled. Is required if evaluate_during_training is enabled.
            verbose (optional): whether output staff.
            **kwargs: Additional metrics that should be used. Pass in the metrics as keyword arguments (name of metric: function to use).
                        A metric function should take in two parameters. The first parameter will be the true labels, and the second parameter will be the predictions. Both inputs
                        will be lists of strings. Note that this will slow down training significantly as the predicted sequences need to be generated.

        Returns: 
        """  # noqa: ignore flake8"

        if args:
            self.args.update_from_dict(args)

        if self.args.evaluate_during_training and eval_data is None:
            raise ValueError(
                "evaluate_during_training is enabled but eval_data is not specified."
                " Pass eval_data to model.train_model() if using evaluate_during_training."
            )

        if not output_dir:
            output_dir = self.args.output_dir

        if os.path.exists(output_dir) and os.listdir(output_dir) and not self.args.overwrite_output_dir:
            raise ValueError(
                "Output directory ({}) already exists and is not empty."
                " Set args.overwrite_output_dir = True to overcome.".format(output_dir)
            )

        train_dataset = self.load_and_cache_examples(train_data, verbose=verbose)

        os.makedirs(output_dir, exist_ok=True)

        os.environ['MASTER_ADDR'] = 'localhost'
        port = random.randint(10000, 20000)
        os.environ['MASTER_PORT'] = str(port)
        mp.spawn(self.train_each_proc, nprocs=self.args.n_gpu,
                 args=(train_dataset, output_dir,
                       show_running_loss, eval_data, verbose, kwargs))

        # self.save_model(model=self.model)

        if verbose:
            logger.info(" Training of {} model complete. Saved to {}.".format(self.args.model_name, output_dir))

    def train_each_proc(self, process_index, train_dataset, *train_args):
        """
        A wrapper function of train() for each process of DDP.
        :param process_index: param train_dataset passed into train().
        :param train_dataset: The training set.
        :param train_args: other position arguments passed to train().
        :return: The same as train().
        """
        self._local_rank = process_index
        self._world_size = self.args.n_gpu
        self.train(train_dataset, *train_args[:-1], **train_args[-1])

    def train(
            self, train_dataset, output_dir, show_running_loss=True, eval_data=None, verbose=True, **kwargs,
    ):
        """
        Trains the model on train_dataset.

        Utility function to be used by the train_model() method. Not intended to be used directly.
        """

        args = self.args
        self.device = torch.device(f"cuda:{self._local_rank}")
        self._move_model_to_device()
        torch.distributed.init_process_group(
            backend='nccl',
            init_method='env://',
            world_size=self._world_size,
            rank=self._local_rank
        )
        self.model = torch.nn.parallel.DistributedDataParallel(self.model, device_ids=[self._local_rank])
        model = self.model

        if self._local_rank == 0:
            tb_writer = SummaryWriter(logdir=args.tensorboard_dir)
        train_sampler = DistributedSampler(
            train_dataset,
            num_replicas=self._world_size,
            rank=self._local_rank
        )
        train_dataloader = DataLoader(
            train_dataset,
            sampler=train_sampler,
            batch_size=args.train_batch_size // self._world_size,
            pin_memory=True
        )

        if args.max_steps > 0:
            t_total = args.max_steps
            args.num_train_epochs = args.max_steps // (len(train_dataloader) // args.gradient_accumulation_steps) + 1
        else:
            t_total = len(train_dataloader) // args.gradient_accumulation_steps * args.num_train_epochs

        no_decay = ["bias", "LayerNorm.weight"]

        optimizer_grouped_parameters = []
        custom_parameter_names = set()
        for group in self.args.custom_parameter_groups:
            params = group.pop("params")
            custom_parameter_names.update(params)
            param_group = {**group}
            param_group["params"] = [p for n, p in model.named_parameters() if n in params]
            optimizer_grouped_parameters.append(param_group)

        for group in self.args.custom_layer_parameters:
            layer_number = group.pop("layer")
            layer = f"layer.{layer_number}."
            group_d = {**group}
            group_nd = {**group}
            group_nd["weight_decay"] = 0.0
            params_d = []
            params_nd = []
            for n, p in model.named_parameters():
                if n not in custom_parameter_names and layer in n:
                    if any(nd in n for nd in no_decay):
                        params_nd.append(p)
                    else:
                        params_d.append(p)
                    custom_parameter_names.add(n)
            group_d["params"] = params_d
            group_nd["params"] = params_nd

            optimizer_grouped_parameters.append(group_d)
            optimizer_grouped_parameters.append(group_nd)

        if not self.args.train_custom_parameters_only:
            optimizer_grouped_parameters.extend(
                [
                    {
                        "params": [
                            p
                            for n, p in model.named_parameters()
                            if n not in custom_parameter_names and not any(nd in n for nd in no_decay)
                        ],
                        "weight_decay": args.weight_decay,
                    },
                    {
                        "params": [
                            p
                            for n, p in model.named_parameters()
                            if n not in custom_parameter_names and any(nd in n for nd in no_decay)
                        ],
                        "weight_decay": 0.0,
                    },
                ]
            )

        warmup_steps = math.ceil(t_total * args.warmup_ratio)
        args.warmup_steps = warmup_steps if args.warmup_steps == 0 else args.warmup_steps

        if 0 < args.save_after < 1:
            args.save_after = math.ceil(t_total * args.save_after)

        if args.optimizer == "AdamW":
            optimizer = AdamW(optimizer_grouped_parameters, lr=args.learning_rate, eps=args.adam_epsilon)
        elif args.optimizer == "Adafactor":
            optimizer = Adafactor(
                optimizer_grouped_parameters,
                lr=args.learning_rate,
                eps=args.adafactor_eps,
                clip_threshold=args.adafactor_clip_threshold,
                decay_rate=args.adafactor_decay_rate,
                beta1=args.adafactor_beta1,
                weight_decay=args.weight_decay,
                scale_parameter=args.adafactor_scale_parameter,
                relative_step=args.adafactor_relative_step,
                warmup_init=args.adafactor_warmup_init,
            )
            if self._local_rank == 0:
                print("Using Adafactor for T5")
        else:
            raise ValueError(
                "{} is not a valid optimizer class. Please use one of ('AdamW', 'Adafactor') instead.".format(
                    args.optimizer
                )
            )

        if args.scheduler == "constant_schedule":
            scheduler = get_constant_schedule(optimizer)

        elif args.scheduler == "constant_schedule_with_warmup":
            scheduler = get_constant_schedule_with_warmup(optimizer, num_warmup_steps=args.warmup_steps)

        elif args.scheduler == "linear_schedule_with_warmup":
            scheduler = get_linear_schedule_with_warmup(
                optimizer, num_warmup_steps=args.warmup_steps, num_training_steps=t_total
            )

        elif args.scheduler == "cosine_schedule_with_warmup":
            scheduler = get_cosine_schedule_with_warmup(
                optimizer,
                num_warmup_steps=args.warmup_steps,
                num_training_steps=t_total,
                num_cycles=args.cosine_schedule_num_cycles,
            )

        elif args.scheduler == "cosine_with_hard_restarts_schedule_with_warmup":
            scheduler = get_cosine_with_hard_restarts_schedule_with_warmup(
                optimizer,
                num_warmup_steps=args.warmup_steps,
                num_training_steps=t_total,
                num_cycles=args.cosine_schedule_num_cycles,
            )

        elif args.scheduler == "polynomial_decay_schedule_with_warmup":
            scheduler = get_polynomial_decay_schedule_with_warmup(
                optimizer,
                num_warmup_steps=args.warmup_steps,
                num_training_steps=t_total,
                lr_end=args.polynomial_decay_schedule_lr_end,
                power=args.polynomial_decay_schedule_power,
            )

        else:
            raise ValueError("{} is not a valid scheduler.".format(args.scheduler))

        if (
                args.model_name
                and os.path.isfile(os.path.join(args.model_name, "optimizer.pt"))
                and os.path.isfile(os.path.join(args.model_name, "scheduler.pt"))
        ):
            # Load in optimizer and scheduler states
            optimizer.load_state_dict(torch.load(os.path.join(args.model_name, "optimizer.pt")))
            scheduler.load_state_dict(torch.load(os.path.join(args.model_name, "scheduler.pt")))

        if self._local_rank == 0:
            logger.info(" Training started")

        global_step = 0
        training_progress_scores = None
        tr_loss, logging_loss = 0.0, 0.0
        model.zero_grad()
        train_iterator = trange(int(args.num_train_epochs), desc="Epoch",
                                disable=args.silent or self._local_rank != 0, mininterval=0)
        epoch_number = 0
        best_eval_metric = None
        current_loss = None
        early_stopping_counter = 0
        steps_trained_in_current_epoch = 0
        epochs_trained = 0
        stop_training = False

        if args.model_name and os.path.exists(args.model_name):
            try:
                # set global_step to global_step of last saved checkpoint from model path
                checkpoint_suffix = args.model_name.split("/")[-1].split("-")
                if len(checkpoint_suffix) > 2:
                    checkpoint_suffix = checkpoint_suffix[1]
                else:
                    checkpoint_suffix = checkpoint_suffix[-1]
                global_step = int(checkpoint_suffix)
                epochs_trained = global_step // (len(train_dataloader) // args.gradient_accumulation_steps)
                steps_trained_in_current_epoch = global_step % (
                        len(train_dataloader) // args.gradient_accumulation_steps
                )

                logger.info("   Continuing training from checkpoint, will skip to saved global_step")
                logger.info("   Continuing training from epoch %d", epochs_trained)
                logger.info("   Continuing training from global step %d", global_step)
                logger.info("   Will skip the first %d steps in the current epoch", steps_trained_in_current_epoch)
            except ValueError:
                logger.info("   Starting fine-tuning.")

        if args.evaluate_during_training:
            training_progress_scores = self._create_training_progress_scores(**kwargs)

        if args.wandb_project and self._local_rank == 0:
            wandb.init(project=args.wandb_project, config={**asdict(args)}, **args.wandb_kwargs)
            wandb.watch(self.model)

        if args.fp16:
            from torch.cuda import amp

            scaler = amp.GradScaler()

        for epoch in train_iterator:
            model.train()
            train_sampler.set_epoch(epoch)
            if epochs_trained > 0:
                epochs_trained -= 1
                continue
            if self._local_rank == 0:
                train_iterator.set_description(f"Epoch {epoch_number + 1} of {args.num_train_epochs}")
            batch_iterator = tqdm(
                train_dataloader,
                desc=f"Running Epoch {epoch_number} of {args.num_train_epochs} on process {self._local_rank}",
                disable=args.silent or self._local_rank != 0,
                mininterval=0,
            )
            for step, batch in enumerate(batch_iterator):
                if steps_trained_in_current_epoch > 0:
                    steps_trained_in_current_epoch -= 1
                    continue

                inputs = self._get_inputs_dict(batch)
                if args.fp16:
                    with amp.autocast():
                        loss = self.compute_loss(model, args, inputs)
                else:
                    loss = self.compute_loss(model, args, inputs)

                loss_ = loss.clone()
                torch.distributed.barrier()
                torch.distributed.reduce(loss_, 0)
                current_loss = loss_.item() / self._world_size

                if show_running_loss and self._local_rank == 0:
                    batch_iterator.set_description(
                        f"Epochs {epoch_number}/{args.num_train_epochs}. Running Loss: {current_loss:9.4f}"
                    )

                if args.gradient_accumulation_steps > 1:
                    loss = loss / args.gradient_accumulation_steps

                if args.fp16:
                    scaler.scale(loss).backward()
                else:
                    loss.backward()

                tr_loss += loss.item()
                if (step + 1) % args.gradient_accumulation_steps == 0:
                    if args.fp16:
                        scaler.unscale_(optimizer)
                    if args.optimizer == "AdamW":
                        torch.nn.utils.clip_grad_norm_(model.parameters(), args.max_grad_norm)

                    if args.fp16:
                        scaler.step(optimizer)
                        scaler.update()
                    else:
                        optimizer.step()
                    scheduler.step()  # Update learning rate schedule
                    model.zero_grad()
                    global_step += 1

                    if args.logging_steps > 0 and global_step % args.logging_steps == 0 and self._local_rank == 0:
                        # Log metrics
                        tb_writer.add_scalar("lr", scheduler.get_last_lr()[0], global_step)
                        tb_writer.add_scalar("loss", (tr_loss - logging_loss) / args.logging_steps, global_step)
                        logging_loss = tr_loss
                        if args.wandb_project or self.is_sweeping:
                            wandb.log(
                                {
                                    "Training loss": current_loss,
                                    "lr": scheduler.get_last_lr()[0]
                                },
                                step=global_step
                            )

                    if args.save_steps > 0 and global_step % args.save_steps == 0 and self._local_rank == 0:
                        # Save model checkpoint
                        output_dir_current = os.path.join(output_dir, "checkpoint-{}".format(global_step))

                        self.save_model(output_dir_current, optimizer, scheduler, model=model)

                    if args.evaluate_during_training and (
                            args.evaluate_during_training_steps > 0
                            and global_step % args.evaluate_during_training_steps == 0
                    ):
                        results = self.eval_model(
                            eval_data,
                            verbose=verbose and args.evaluate_during_training_verbose,
                            silent=args.evaluate_during_training_silent or self._local_rank != 0,
                            **kwargs,
                        )

                        if self._local_rank == 0:
                            for key, value in results.items():
                                tb_writer.add_scalar("eval_{}".format(key), value, global_step)

                            output_dir_current = os.path.join(output_dir, "checkpoint-{}".format(global_step))

                            if args.save_eval_checkpoints:
                                self.save_model(output_dir_current, optimizer, scheduler, model=model, results=results)

                            stop_training, best_eval_metric, early_stopping_counter = self.logging_and_saving(
                                args,
                                results,
                                global_step,
                                train_iterator,
                                optimizer,
                                scheduler,
                                model,
                                training_progress_scores,
                                current_loss,
                                best_eval_metric,
                                verbose,
                                early_stopping_counter)

                        torch.distributed.barrier()
                        stop_training_tensor = torch.tensor([stop_training], device=self.device)
                        torch.distributed.broadcast(stop_training_tensor, src=0)
                        stop_training = bool(stop_training_tensor.cpu()[0])
                        if stop_training:
                            break

                        model.train()

            if stop_training:
                break
            epoch_number += 1
            output_dir_current = os.path.join(output_dir, "checkpoint-{}-epoch-{}".format(global_step, epoch_number))

            if (args.save_model_every_epoch or args.evaluate_during_training) and self._local_rank == 0:
                os.makedirs(output_dir_current, exist_ok=True)

            if args.save_model_every_epoch and self._local_rank == 0:
                self.save_model(output_dir_current, optimizer, scheduler, model=model)

            if args.evaluate_during_training and args.evaluate_each_epoch:
                results = self.eval_model(
                    eval_data,
                    verbose=verbose and args.evaluate_during_training_verbose,
                    silent=args.evaluate_during_training_silent or self._local_rank != 0,
                    **kwargs,
                )

                if self._local_rank == 0:
                    if args.save_eval_checkpoints:
                        self.save_model(output_dir_current, optimizer, scheduler, results=results)

                    stop_training, best_eval_metric, early_stopping_counter = self.logging_and_saving(
                        args,
                        results,
                        global_step,
                        train_iterator,
                        optimizer,
                        scheduler,
                        model,
                        training_progress_scores,
                        current_loss,
                        best_eval_metric,
                        verbose,
                        early_stopping_counter)

                torch.distributed.barrier()
                stop_training_tensor = torch.tensor([stop_training], device=self.device)
                torch.distributed.broadcast(stop_training_tensor, src=0)
                stop_training = bool(stop_training_tensor.cpu()[0])
                if stop_training:
                    break

        # close tensorboard writer to avoid EOFError.
        if self._local_rank == 0:
            tb_writer.close()
            wandb.finish()

    def eval_model(
            self, eval_data, output_dir=None, verbose=True, silent=False, **kwargs
    ):
        """
        Evaluates the model on eval_data. Saves results to output_dir.
        Args:
            eval_data: Pandas DataFrame containing the 3 columns - `prefix`, `input_text`, `target_text`.
                        - `prefix`: A string indicating the task to perform. (E.g. `"question"`, `"stsb"`)
                        - `input_text`: The input text sequence. `prefix` is automatically prepended to form the full input. (<prefix>: <input_text>)
                        - `target_text`: The target sequence
            output_dir: The directory where model files will be saved. If not given, self.args.output_dir will be used.
            verbose: If verbose, results will be printed to the console on completion of evaluation.
            silent: If silent, tqdm progress bars will be hidden.
            **kwargs: Additional metrics that should be used. Pass in the metrics as keyword arguments (name of metric: function to use).
                        A metric function should take in two parameters. The first parameter will be the true labels, and the second parameter will be the predictions. Both inputs
                        will be lists of strings. Note that this will slow down evaluation significantly as the predicted sequences need to be generated.
        Returns:
            results: Dictionary containing evaluation results.
        """  # noqa: ignore flake8"

        if not output_dir:
            output_dir = self.args.output_dir

        eval_dataset = self.load_and_cache_examples(
            eval_data, evaluate=True, verbose=verbose, silent=silent
        )
        os.makedirs(output_dir, exist_ok=True)

        result = self.evaluate(
            eval_dataset, output_dir, verbose=verbose, silent=silent, **kwargs
        )
        self.results.update(result)

        if self.args.evaluate_generated_text:
            if self.args.preprocess_inputs:
                to_predict = [
                    prefix + ": " + input_text
                    for prefix, input_text in zip(
                        eval_data["prefix"], eval_data["input_text"]
                    )
                ]
            else:
                to_predict = [
                    prefix + input_text
                    for prefix, input_text in zip(
                        eval_data["prefix"], eval_data["input_text"]
                    )
                ]
            preds = self.predict(to_predict)

            result = self.compute_metrics(
                eval_data["target_text"].tolist(), preds, **kwargs
            )
            self.results.update(result)

        if verbose:
            logger.info(self.results)

        return self.results

    def evaluate(self, eval_dataset, output_dir, verbose=True, silent=False, **kwargs):
        """
        Evaluates the model on eval_dataset.
        Utility function to be used by the eval_model() method. Not intended to be used directly.
        """

        model = self.model
        args = self.args
        eval_output_dir = output_dir

        results = {}
        eval_sampler = DistributedSampler(
            eval_dataset,
            num_replicas=self._world_size,
            rank=self._local_rank
        )
        eval_dataloader = DataLoader(
            eval_dataset,
            sampler=eval_sampler,
            batch_size=args.eval_batch_size // self._world_size,
            pin_memory=True
        )

        eval_loss = 0.0
        nb_eval_steps = 0
        model.eval()

        if self.args.fp16:
            from torch.cuda import amp

        for batch in tqdm(
                eval_dataloader,
                disable=args.silent or silent,
                desc="Running Evaluation"
        ):
            inputs = self._get_inputs_dict(batch)
            with torch.no_grad():
                if self.args.fp16:
                    with amp.autocast():
                        outputs = model(**inputs)
                        loss = outputs[0]
                else:
                    outputs = model(**inputs)
                    loss = outputs[0]
                torch.distributed.barrier()
                torch.distributed.reduce(loss, 0)
                eval_loss += loss.item()
            nb_eval_steps += 1

        eval_loss = eval_loss / nb_eval_steps / self._world_size

        if self._local_rank == 0:
            print(eval_loss)

        results["eval_loss"] = eval_loss

        if self._local_rank == 0:
            output_eval_file = os.path.join(eval_output_dir, "eval_results.txt")
            with open(output_eval_file, "w") as writer:
                for key in sorted(results.keys()):
                    writer.write("{} = {}\n".format(key, str(results[key])))

        return results

    def logging_and_saving(
            self,
            args,
            results,
            global_step,
            train_iterator,
            optimizer,
            scheduler,
            model,
            training_progress_scores,
            current_loss,
            best_eval_metric,
            verbose,
            early_stopping_counter):
        training_progress_scores["global_step"].append(global_step)
        training_progress_scores["train_loss"].append(current_loss)
        for key in results:
            training_progress_scores[key].append(results[key])
        report = pd.DataFrame(training_progress_scores)
        report.to_csv(
            os.path.join(args.output_dir, "training_progress_scores.csv"), index=False,
        )

        if args.wandb_project or self.is_sweeping:
            wandb.log(self._get_last_metrics(training_progress_scores), step=global_step)

        stop_training = False
        if global_step > args.save_after:
            if not best_eval_metric:
                best_eval_metric = results[args.early_stopping_metric]
                self.save_model(args.best_model_dir, optimizer, scheduler, model=model, results=results)

            if args.early_stopping_metric_minimize:
                if results[args.early_stopping_metric] - best_eval_metric < args.early_stopping_delta:
                    best_eval_metric = results[args.early_stopping_metric]
                    self.save_model(args.best_model_dir, optimizer, scheduler, model=model, results=results)
                    early_stopping_counter = 0
                else:
                    stop_training, early_stopping_counter = \
                        self.check_early_stopping(early_stopping_counter, args, train_iterator, verbose)
            else:
                if results[args.early_stopping_metric] - best_eval_metric > args.early_stopping_delta:
                    best_eval_metric = results[args.early_stopping_metric]
                    self.save_model(args.best_model_dir, optimizer, scheduler, model=model, results=results)
                    early_stopping_counter = 0
                else:
                    stop_training, early_stopping_counter = \
                        self.check_early_stopping(early_stopping_counter, args, train_iterator, verbose)

        return stop_training, best_eval_metric, early_stopping_counter

    def check_early_stopping(self, early_stopping_counter, args, train_iterator, verbose):
        stop_training = False
        if args.use_early_stopping:
            if early_stopping_counter < args.early_stopping_patience:
                early_stopping_counter += 1
                if verbose:
                    logger.info(f" No improvement in {args.early_stopping_metric}")
                    logger.info(f" Current step: {early_stopping_counter}")
                    logger.info(f" Early stopping patience: {args.early_stopping_patience}")
            else:
                if verbose:
                    logger.info(f" Patience of {args.early_stopping_patience} steps reached")
                    logger.info(" Training terminated.")
                    train_iterator.close()
                stop_training = True
        return stop_training, early_stopping_counter

    def compute_loss(self, model, args, inputs):
        outputs = model(**inputs)
        if args.r_drop:
            outputs_ = model(**inputs)
            loss = self.compute_r_drop_loss(
                outputs['loss'],
                outputs_['loss'],
                outputs['logits'],
                outputs_['logits'],
                inputs['attention_mask'],
                args.r_drop_alpha
            )
        else:
            loss = outputs[0]
        return loss

    def compute_kl_loss(self, p, q, pad_mask=None, reduction='mean'):

        p_loss = F.kl_div(F.log_softmax(p, dim=-1), F.softmax(q, dim=-1), reduction='none')
        q_loss = F.kl_div(F.log_softmax(q, dim=-1), F.softmax(p, dim=-1), reduction='none')

        if pad_mask is not None:
            p_loss.masked_fill_(pad_mask.to(bool).unsqueeze(-1), 0.)
            q_loss.masked_fill_(pad_mask.to(bool).unsqueeze(-1), 0.)

        if reduction == 'mean':
            p_loss = p_loss.mean()
            q_loss = q_loss.mean()
        elif reduction == 'sum':
            p_loss = p_loss.sum()
            q_loss = q_loss.sum()
        else:
            raise ValueError('Only mean or sum reduction is supported in computing KL Divergence!')

        loss = (p_loss + q_loss) / 2
        return loss

    def compute_r_drop_loss(self, ce1, ce2, logit1, logit2, attention_mask, alpha, reduction='mean'):
        kl_loss = self.compute_kl_loss(logit1, logit2, attention_mask, reduction=reduction)
        ce_loss = 0.5 * (ce1 + ce2)
        return ce_loss + alpha * kl_loss



# encoding: utf-8
"""
@author: zyl
@file: data_processing.py
@time: 2021/11/25 14:15
@desc:
"""

import re


class DTProcesser:
    def __init__(self):
        pass

    @staticmethod
    def cut_train_eval(all_df,max_sample_num=2000):
        from sklearn.utils import resample
        raw_df = resample(all_df, replace=False)
        cut_point = min(max_sample_num, int(0.2 * len(raw_df)))
        eval_df = raw_df[0:cut_point]
        train_df = raw_df[cut_point:]
        return train_df, eval_df

    @staticmethod
    def remove_illegal_chars(text):
        ILLEGAL_CHARACTERS_RE = re.compile(r'[\000-\010]|[\013-\014]|[\016-\037]')
        return ILLEGAL_CHARACTERS_RE.sub(r'', str(text))  # 非法字符

    @staticmethod
    def _remove_invisible_chars(s, including_char=('\t', '\n', '\r')):
        """移除所有不可见字符，除'\t', '\n', '\r'外"""
        str = ''
        for x in s:
            if (x not in including_char) and (not x.isprintable()):
                str += ' '
            else:
                str += x
        return str



# encoding: utf-8
'''
@author: zyl
@file: Analyzer.py
@time: 2021/11/11 9:34
@desc:
'''
import langid
import matplotlib.pyplot as plt

class BaseAnalyzer:
    def __init__(self):
        pass

    def run(self):
        pass

    @staticmethod
    def get_text_string_length(text:str):
        return len(text)

    @staticmethod
    def get_text_word_length(text: str):
        if langid.classify(text)[0]=='zh':
            return len(text)  # zh -word-piece
        else:
            return len(text.split())  # en - split by space

    @staticmethod
    def get_text_token_length(text: str, tokenizer=None):
        if not tokenizer:
            from transformers import BertTokenizer
            tokenizer = BertTokenizer.from_pretrained('bert-base-multilingual-cased')
        return len(tokenizer.tokenize(text))

    @staticmethod
    def show_df_base_info(df):
        desc = df.describe(percentiles=[0.10,0.25,0.5,0.75,0.85,0.95,0.99])
        print(desc)
        info = df.info()
        print(info)

    @staticmethod
    def draw_box(df,column):
        plt.subplots()
        plt.boxplot(df[column])
        plt.show()

    @staticmethod
    def draw_hist(df,column):
        # df['input_text_length'].plot.hist(bins=2000)  # 列的直方图
        plt.subplots()
        plt.hist(df[column])
        plt.show()

if __name__ == '__main__':

    pass



# encoding: utf-8
"""
@author: zyl
@file: ner_model.py
@time: 2021/11/25 13:59
@desc:
"""
import time

import pandas as pd
import wandb
from loguru import logger
from simpletransformers.ner import NERModel


class NerModel:
    """
    ner model for train and eval
    """

    def __init__(self):
        self.start_time = '...'
        self.end_time = '...'
        self.describe = " use simple-transformers--ner-model"
        self.show_running_loss = False

        self.wandb_proj = 'ner'
        self.save_dir = './'
        self.model_version = 'v0.0.0.0'  # to save model or best model
        # like a,b,c,d : a 原始数据批次，b模型方法批次，比如mt5和分类，
        # c进行模型的处理的数据批次，比如同一输入，输出是文本还是序号，d：迭代调参批次

        self.model_type = 'roberta'
        self.pretrained_model = 'roberta-base'  # 预训练模型位置 model_name

        self.use_cuda = True
        self.cuda_device = 0
        self.labels = ["O", "B-MISC", "I-MISC", "B-PER", "I-PER", "B-ORG", "I-ORG", "B-LOC", "I-LOC"]

        self.model_args = self.my_config()

    def my_config(self):
        return {
            'train_batch_size': 8,

            'use_multiprocessing': False,
            'use_multiprocessing_for_evaluation': False,
            # multiprocess

            # base config
            'reprocess_input_data': True,
            'use_cached_eval_features': False,
            'fp16': False,
            'manual_seed': 234,
            'gradient_accumulation_steps': 1,  # ::increase batch size,Use time for memory,

            # save
            'no_save': False,
            'save_eval_checkpoints': False,
            'save_model_every_epoch': False,
            'save_optimizer_and_scheduler': True,
            'save_steps': -1,

            # eval
            'evaluate_during_training': True,
            'evaluate_during_training_verbose': True,

            'no_cache': False,
            'use_early_stopping': False,
            'encoding': None,
            'do_lower_case': False,
            'dynamic_quantize': False,
            'quantized_model': False,
            'silent': False,

            'overwrite_output_dir': True,
            'output_dir': self.save_dir + 'outputs/' + self.model_version + '/',
            'cache_dir': self.save_dir + 'cache/' + self.model_version + '/',
            'best_model_dir': self.save_dir + 'best_model/' + self.model_version + '/',
            'tensorboard_dir': self.save_dir + 'runs/' + self.model_version + '/' + time.strftime("%Y%m%d_%H%M%S",
                                                                                                  time.localtime()) + '/',
        }

    @staticmethod
    def deal_with_df(df):
        df = df[["sentence_id", "words", "labels"]]
        df = df.astype({'sentence_id': 'int', 'words': 'str', 'labels': 'str'})
        return df

    def train(self, train_data: pd.DataFrame, eval_data: pd.DataFrame):
        # deal with dt
        train_data = NerModel.deal_with_df(train_data)
        eval_data = NerModel.deal_with_df(eval_data)
        train_size = len(set(train_data['sentence_id'].tolist()))
        eval_size = len(set(eval_data['sentence_id'].tolist()))

        all_steps = train_size / self.model_args.get('train_batch_size')
        self.model_args.update(
            {
                'train_size': train_size,
                'eval_size': eval_size,
                'logging_steps': int(max(all_steps / 10 / self.model_args.get('gradient_accumulation_steps'), 1)),
                'evaluate_during_training_steps': int(
                    max(all_steps / 10 / self.model_args.get('gradient_accumulation_steps'), 1)),
                'wandb_project': self.wandb_proj,
                'wandb_kwargs': {
                    'name': self.model_version + time.strftime("_%m%d_%H:%M:%S", time.localtime()),
                    'tags': [self.model_version, 'train']
                }
            }
        )

        # get model
        model = NERModel(model_type=self.model_type, model_name=self.pretrained_model, labels=self.labels,
                         args=self.model_args, use_cuda=self.use_cuda, cuda_device=self.cuda_device)

        # train
        try:
            start_time = time.time()
            logger.info(f'start training: model_version---{self.model_version}')
            model.train_model(train_data=train_data, eval_data=eval_data)
            logger.info('training finished!!!')
            end_time = time.time()
            logger.info(f'train time: {round(end_time - start_time, 4)} s')
        except Exception as error:
            logger.error(f'train failed!!! ERROR:{error}')
        finally:
            wandb.finish()
            # ModelUtils.remove_some_model_files(model.args)

    def train_example(self):
        train_file = './test.xlsx'
        eval_file = './test.xlsx'
        train_data = pd.read_excel(train_file)
        eval_data = pd.read_excel(eval_file)

        self.save_dir = './'
        self.model_version = 'erv4.2.0.2'
        self.model_type = 'bert'
        self.pretrained_model = 'bert-base-multilingual-cased'  # 预训练模型位置 model_name
        self.use_cuda = True
        self.cuda_device = 0
        self.labels = ["O", "B-DISEASE", "I-DISEASE"]

        self.model_args = self.my_config()
        self.model_args.update(
            {
                'train_file': train_file,
                'eval_file': eval_file,
                'num_train_epochs': 3,
                'learning_rate': 1e-3,
                'train_batch_size': 24,  # 28
                'gradient_accumulation_steps': 16,
                'eval_batch_size': 16,
                'max_seq_length': 512,
            }
        )
        self.train(train_data,eval_data)

    @staticmethod
    def eval_decoration(eval_func):
        # #############################################################
        # examples: should set : self.wandb_proj , self.ver , self.args.hyper_args
        # >>> @eval_decoration
        # >>> def eval(eval_df,a,b):
        # >>>     eval_res = func... a,b
        # >>>     return eval_res
        # ############################################################
        def eval_method(self, eval_df, *args, **kwargs):
            evel_size = self.model_args.get('eval_size')
            # wand_b
            wandb.init(project=self.wandb_proj, config=self.model_args,
                       name=self.model_version + time.strftime("_%m%d_%H:%M:%S", time.localtime()),
                       tags=[self.model_version, 'eval'])
            try:
                start_time = time.time()
                logger.info(f'start eval: model_version---{self.model_version},eval size---{evel_size}')
                eval_res = eval_func(self, eval_df, *args, **kwargs)  # type:dict
                logger.info('eval finished!!!')
                end_time = time.time()
                need_time = round((end_time - start_time) / evel_size, 5)
                eval_time = round(need_time * evel_size, 4)
                print(f'eval results: {eval_res}')
                logger.info(f'eval time: {need_time} s * {evel_size} = {eval_time} s')
                assert isinstance(eval_res, dict) == True
                eval_res.update({"eval_length": evel_size})
                wandb.log(eval_res)
            except Exception as error:
                logger.error(f'eval failed!!! ERROR:{error}')
                eval_res = dict()
            finally:
                wandb.finish()
            return eval_res

        return eval_method

    @staticmethod
    def get_entity(pred_list, label='DISEASE'):
        if not label:
            label = ''
        entities = []
        e = ''
        is_entity = 0
        for index, p in enumerate(pred_list):
            if p == '0':
                if is_entity == 1:
                    entities.append(e)
                is_entity = 0
            elif p.startswith('B-' + label):
                if is_entity == 1:
                    if e:
                        entities.append(e)
                e = '-' + str(index)
                is_entity = 1
            elif p.startswith('I-' + label):
                e = e + ('-' + str(index))
        if is_entity == 1:
            entities.append(e)
        return entities


    def eval(self, eval_df: pd.DataFrame):
        eval_data = NerModel.deal_with_df(eval_df)
        eval_size = len(set(eval_df['sentence_id'].tolist()))

        self.model_args.update(
            {
                'eval_size': eval_size,
                'wandb_project': self.wandb_proj,
                'wandb_kwargs': {
                    'name': self.model_version + time.strftime("_%m%d_%H:%M:%S", time.localtime()),
                    'tags': [self.model_version, 'eval']
                }
            }
        )

        model = NERModel(model_type=self.model_type, model_name=self.model_args.get('best_model_dir'),
                         args=self.model_args, use_cuda=self.use_cuda, cuda_device=self.cuda_device)

        result, model_outputs, preds_list = model.eval_model(eval_data)

        wandb.init(
            project=self.wandb_proj,
            config = self.model_args,
            name=self.model_version + time.strftime("_%m%d_%H:%M:%S", time.localtime()),
            tags=[self.model_version, 'eval']
        )
        wandb.log({"f1_score": result.get('f1_score')})

        return result


    def eval_sample(self):
        eval_file = './test.xlsx'
        eval_data = pd.read_excel(eval_file)

        self.save_dir = './'
        self.model_version = 'erv4.2.0.2'
        self.model_type = 'bert'
        self.use_cuda = True
        self.cuda_device = 1

        self.model_args = self.my_config()
        self.model_args.update(
            {
                'eval_file': eval_file,
                'eval_batch_size': 16,
                'max_seq_length': 512,
            }
        )
        self.eval(eval_data)

if __name__ == '__main__':
    s = ['O', 'O', 'O', 'B-DISEASE', 'I-DISEASE', 'O', 'B-DISEASE', 'B-DISEASE', 'B-DISEASE', 'I-DISEASE',
         'I-DISEASE', 'O', 'B-DISEASE', 'O', 'I-DISEASE', 'I-DISEASE', 'B-DISEASE', 'I-DISEASE']

    print(NerModel.get_entity(s))



# encoding: utf-8
'''
@author: zyl
@file: my_model.py
@time: 2021/11/11 10:56
@desc:
'''

import time

import pandas as pd
import wandb
from loguru import logger
from simpletransformers.classification import ClassificationModel, ClassificationArgs, DDPClassificationModel
from simpletransformers.t5 import T5Args

from zyl_utils.model_utils.my_T5model import MyT5, MyDDPT5


class MyModel:
    """
    my model for train and eval
    """

    def __init__(self):
        self.start_time = '...'
        self.end_time = '...'

        self.wandb_proj = 'test'
        self.model_version = 'test'  # to save model or best model
        # like a,b,c,d : a 原始数据批次，b模型方法批次，比如mt5和分类，
        # c进行模型的数据批次，比如同一输入，输出是文本还是序号，d：迭代调参批次

        self.use_model = 'classification'  # mt5 /classification
        self.model_type = 'bert'
        self.pretrained_model = './best/v1.1.1.1/'  # 预训练模型位置

        self.use_cuda = True
        self.cuda_device = 0
        self.num_labels = 2

        self.args = MyModel.set_model_parameter(model_version=self.model_version,
                                                args=self._set_args(), save_dir='./')

    def _set_args(self):
        if self.use_model == 't5' or self.use_model == 'mt5':
            return T5Args()
        else:
            return ClassificationArgs()

    @staticmethod
    def set_model_parameter(model_version='test', args=ClassificationArgs(), save_dir='./'):
        # multiprocess
        args.use_multiprocessing = False
        args.use_multiprocessing_for_evaluation = False

        # base config
        args.reprocess_input_data = True
        args.use_cached_eval_features = False
        args.fp16 = False
        args.manual_seed = 234
        args.gradient_accumulation_steps = 2  # ==increase batch size,Use time for memory,

        # save
        args.no_save = False
        args.save_eval_checkpoints = False
        args.save_model_every_epoch = False
        args.save_optimizer_and_scheduler = True
        args.save_steps = -1

        # eval
        args.evaluate_during_training = True
        args.evaluate_during_training_verbose = True

        args.no_cache = False
        args.use_early_stopping = False
        args.encoding = None
        args.do_lower_case = False
        args.dynamic_quantize = False
        args.quantized_model = False
        args.silent = False

        args.overwrite_output_dir = True
        args.output_dir = save_dir + 'outputs/' + model_version + '/'
        args.cache_dir = save_dir + 'cache/' + model_version + '/'
        args.best_model_dir = save_dir + 'best_model/' + model_version + '/'
        args.tensorboard_dir = save_dir + 'runs/' + model_version + '/' + time.strftime("%Y%m%d_%H%M%S",
                                                                                        time.localtime()) + '/'
        return args

    def get_train_model(self):
        if self.args.n_gpu <= 1:
            if self.use_model == 't5' or self.use_model == 'mt5':
                self.args.use_multiprocessed_decoding = False
                return MyT5(model_type=self.model_type, model_name=self.pretrained_model,
                            use_cuda=self.use_cuda, cuda_device=self.cuda_device, args=self.args)
            else:
                return ClassificationModel(model_type=self.model_type, model_name=self.pretrained_model,
                                           use_cuda=self.use_cuda, cuda_device=self.cuda_device, args=self.args,
                                           num_labels=self.num_labels)
        else:
            if self.use_model == 't5' or self.use_model == 'mt5':
                self.args.use_multiprocessed_decoding = False
                return MyDDPT5(model_type=self.model_type, model_name=self.pretrained_model, use_cuda=True,
                               cuda_device=-1, args=self.args)
            elif self.use_model == 'classification':
                return ClassificationModel(model_type=self.model_type, model_name=self.pretrained_model,
                                           use_cuda=self.use_cuda, cuda_device=self.cuda_device, args=self.args,
                                           num_labels=self.num_labels)
            else:
                return DDPClassificationModel(model_type=self.model_type, model_name=self.pretrained_model,
                                              use_cuda=True, args=self.args, num_labels=self.num_labels)

    @staticmethod
    def deal_with_df(df, use_model='cls'):
        if use_model == 't5' or use_model == 'mt5':
            df = df[['prefix', 'input_text', 'target_text']]
            df = df.astype('str')
        elif use_model == 'sentence_pair':
            df = df[['text_a', 'text_b', 'labels']]
            df = df.astype({'text_a': 'str', 'text_b': 'str', 'labels': 'int'})
        else:
            df = df.astype({'text': 'str', 'labels': 'int'})
            df = df[['text', 'labels']]
        return df

    def train(self, train_df: pd.DataFrame, eval_df: pd.DataFrame, if_send_message=False):
        # deal with dt
        train_df = MyModel.deal_with_df(train_df, use_model=self.use_model)
        eval_df = MyModel.deal_with_df(eval_df, use_model=self.use_model)

        # config some parameters
        train_size = train_df.shape[0]
        self.args.update_from_dict({'train_length': train_size})
        all_steps = train_size / self.args.train_batch_size
        self.args.logging_steps = int(max(all_steps / 10 / self.args.gradient_accumulation_steps, 1))
        self.args.evaluate_during_training_steps = int(
            max(all_steps / 10 / self.args.gradient_accumulation_steps, 1))

        self.args.wandb_project = self.wandb_proj
        self.args.wandb_kwargs = {
            'name': self.model_version + time.strftime("_%m%d_%H:%M:%S", time.localtime()),
            'tags': [self.model_version, 'train']}

        # get model
        model = self.get_train_model()

        # train
        try:
            start_time = time.time()
            logger.info(f'start training: model_version---{self.model_version},train length---{train_size}')
            if self.use_model == 't5' or self.use_model == 'mt5':
                model.train_model(train_data=train_df, eval_data=eval_df)
            else:
                model.train_model(train_df=train_df, eval_df=eval_df)
            logger.info('training finished!!!')
            end_time = time.time()
            logger.info(f'train time: {round(end_time - start_time, 4)} s')
        except Exception as error:
            logger.error(f'train failed!!! ERROR:{error}')
            if if_send_message:
                print(f'train failed!!! ERROR:{error}')
                # ModelUtils.send_to_me(f'train failed!!! ERROR:{error}')
        finally:
            wandb.finish()
            # ModelUtils.remove_some_model_files(model.args)

    def get_predict_model(self):
        if self.args.n_gpu <= 1:
            if self.use_model == 't5' or self.use_model == 'mt5':
                self.args.use_multiprocessed_decoding = False
                return MyT5(model_type=self.model_type, model_name=self.args.best_model_dir,
                            use_cuda=self.use_cuda, cuda_device=self.cuda_device, args=self.args)
            else:
                return ClassificationModel(model_type=self.model_type, model_name=self.args.best_model_dir,
                                           use_cuda=self.use_cuda, cuda_device=self.cuda_device, args=self.args,
                                           num_labels=self.num_labels)
        else:
            if self.use_model == 't5' or self.use_model == 'mt5':
                self.args.use_multiprocessed_decoding = False
                return MyDDPT5(model_type=self.model_type, model_name=self.args.best_model_dir, use_cuda=True,
                               cuda_device=-1, args=self.args)
            elif self.use_model == 'sentence_pair':
                return ClassificationModel(model_type=self.model_type, model_name=self.args.best_model_dir,
                                           use_cuda=self.use_cuda, cuda_device=self.cuda_device, args=self.args,
                                           num_labels=self.num_labels)
            else:
                return DDPClassificationModel(model_type=self.model_type, model_name=self.args.best_model_dir,
                                              use_cuda=True, args=self.args, num_labels=self.num_labels)

    @staticmethod
    def eval_decoration(eval_func):
        # #############################################################
        # examples: should set : self.wandb_proj , self.ver , self.args.hyper_args
        # >>> @eval_decoration
        # >>> def eval(eval_df,a,b):
        # >>>     eval_res = func... a,b
        # >>>     return eval_res
        # ############################################################
        def eval_method(self, eval_df, *args, **kwargs):
            eval_length = eval_df.shape[0]

            # wand_b
            wandb.init(project=self.wandb_proj, config=self.args,
                       name=self.model_version + time.strftime("_%m%d_%H:%M:%S", time.localtime()),
                       tags=[self.model_version, 'eval'])
            try:
                start_time = time.time()
                logger.info(f'start eval: model_version---{self.model_version},eval length---{eval_length}')
                eval_res = eval_func(self, eval_df, *args, **kwargs)  # type:dict
                logger.info('eval finished!!!')
                end_time = time.time()
                need_time = round((end_time - start_time) / eval_length, 5)
                eval_time = round(need_time * eval_length, 4)
                print(f'eval results: {eval_res}')
                logger.info(f'eval time: {need_time} s * {eval_length} = {eval_time} s')
                assert isinstance(eval_res, dict) == True
                eval_res.update({"eval_length": eval_length})
                wandb.log(eval_res)
            except Exception as error:
                logger.error(f'eval failed!!! ERROR:{error}')
                eval_res = dict()
            finally:
                wandb.finish()
            return eval_res

        return eval_method



# encoding: utf-8
"""
@author: zyl
@file: utils.py
@time: 2021/11/29 15:18
@desc:
"""
import time

import pandas as pd
import wandb
from loguru import logger
from simpletransformers.ner import NERModel

class Utils:
    def __init__(self):
        pass

    @staticmethod
    def eval_decoration(eval_func):
        # #############################################################
        # examples: should set : self.wandb_proj , self.ver , self.args.hyper_args
        # >>> @eval_decoration
        # >>> def eval(eval_df,a,b):
        # >>>     eval_res = func... a,b
        # >>>     return eval_res
        # ############################################################
        def eval_method(self, eval_df, *args, **kwargs):
            evel_size = self.model_args.get('eval_size')
            # wand_b
            wandb.init(project=self.wandb_proj, config=self.model_args,
                       name=self.model_version + time.strftime("_%m%d_%H:%M:%S", time.localtime()),
                       tags=[self.model_version, 'eval'])
            try:
                start_time = time.time()
                logger.info(f'start eval: model_version---{self.model_version},eval size---{evel_size}')
                eval_res = eval_func(self, eval_df, *args, **kwargs)  # type:dict
                logger.info('eval finished!!!')
                end_time = time.time()
                need_time = round((end_time - start_time) / evel_size, 5)
                eval_time = round(need_time * evel_size, 4)
                print(f'eval results: {eval_res}')
                logger.info(f'eval time: {need_time} s * {evel_size} = {eval_time} s')
                assert isinstance(eval_res, dict) == True
                eval_res.update({"eval_length": evel_size})
                wandb.log(eval_res)
            except Exception as error:
                logger.error(f'eval failed!!! ERROR:{error}')
                eval_res = dict()
            finally:
                wandb.finish()
            return eval_res

        return eval_method






# encoding: utf-8
"""
@author: zyl
@file: ner_metric.py
@time: 2021/11/29 15:08
@desc:
"""
from typing import List, Set

import pandas as pd


def entity_recognition_v2(y_true: List[Set], y_pred: List[Set], pos_neg_ratio: str = None, self_metric=False):
    """
    the metric of entity_recognition, version-v2, reference: https://docs.qq.com/doc/DYXRYQU1YbkVvT3V2

    Args:
        y_true: list[set],the list of true target texts,each element is a set
        y_pred: list[set],the list of pred target texts,each element is a set
        pos_neg_ratio: the ratio of positive and negative sample importance, default: the ratio of positive and
                       negative sample sizes, you can set it,like"7:3"
        self_metric: self_metric

    Returns:
        show report and res
    """

    neg_data = 0
    neg_correct_dt = 0
    neg_wrong_dt = 0
    neg_redundant_entities = 0

    pos_data = 0
    pos_correct_dt = 0
    pos_wrong_dt = 0
    pos_correct_entities = 0
    pos_wrong_entities = 0
    pos_omitted_entities = 0
    pos_redundant_entities = 0

    for i, j in zip(y_true, y_pred):
        if i == set():
            neg_data += 1
            if j == set():
                neg_correct_dt += 1
            else:
                neg_wrong_dt += 1
                neg_redundant_entities += len(j)
        else:
            pos_data += 1
            true_pred = len(i & j)
            pos_correct_entities += true_pred

            if i == j:
                pos_correct_dt += 1
            elif len(i) >= len(j):
                pos_wrong_dt += 1
                pos_wrong_entities += (len(j) - true_pred)
                pos_omitted_entities += (len(i) - len(j))
            else:
                pos_wrong_dt += 1
                pos_redundant_entities += (len(j) - len(i))
                pos_wrong_entities += (len(i) - true_pred)

    all_pos_entities = pos_correct_entities + pos_wrong_entities + pos_omitted_entities + pos_redundant_entities
    if neg_data == 0:
        neg_metric = 0
    else:
        neg_metric = neg_correct_dt / (neg_correct_dt + neg_redundant_entities)
    if pos_data == 0:
        pos_metric = 0
    else:
        pos_metric = pos_correct_entities / all_pos_entities

    sum_metric_micro = (pos_correct_entities + neg_correct_dt) / (
            neg_correct_dt + neg_redundant_entities + all_pos_entities)
    # sum_metric_macro = neg_metric * 0.5 + pos_metric * 0.5

    if pos_neg_ratio:
        pos_all = float(pos_neg_ratio.split(':')[0])
        neg_all = float(pos_neg_ratio.split(':')[1])
        pos_ratio = pos_all / (pos_all + neg_all)
        neg_ratio = neg_all / (pos_all + neg_all)
    else:
        pos_ratio = pos_data / (pos_data + neg_data)
        neg_ratio = neg_data / (pos_data + neg_data)

    sum_metric_weighted = pos_ratio * pos_metric + neg_ratio * neg_metric

    # pos_precision = pos_correct_dt / (neg_correct_dt + pos_correct_dt)
    # recall = pos_correct_dt / pos_data
    tp = pos_correct_dt
    fn = pos_wrong_dt
    fp = neg_wrong_dt
    tn = neg_correct_dt

    accuracy = (tp + tn) / (tp + fn + fp + tn)
    # precision = tp / (tp + fp)
    # recall = tp / (tp + fn)
    # f1 = 2 / (1 / precision + 1 / recall)
    r = {
        'positive data': [str(pos_data), pos_correct_dt, pos_wrong_dt, pos_correct_entities,
                          pos_wrong_entities, pos_omitted_entities, pos_redundant_entities, pos_metric],
        'negative data': [neg_data, neg_correct_dt, neg_wrong_dt, '-', '-', '-', neg_redundant_entities,
                          neg_metric],

        'all data ': [str(pos_data + neg_data), neg_correct_dt + pos_correct_dt, neg_wrong_dt + pos_wrong_dt,
                      pos_correct_entities, pos_wrong_entities, pos_omitted_entities,
                      pos_redundant_entities + neg_redundant_entities,
                      sum_metric_micro],
        # 'precision': ['', '', '', '', '', '', '', precision],
        # 'recall': ['', '', '', '', '', '', '', recall],
        # 'f1 score': ['', '', '', '', '', '', '', (2 * precision * recall) / (precision + recall)],
        # 'accuracy score': ['', '', '', '', '', '', '', (neg_correct_dt + pos_correct_dt) / (pos_data + neg_data)],
        # 'micro score': ['', '', '', '', '', '', '', sum_metric_micro],
        # 'macro score': ['', '', '', '', '', '', '', sum_metric_macro],
        'weighted score': ['', '', '', '', '', '', '', sum_metric_weighted],
    }

    index = ['| data_num', '| correct_data', '| wrong_data', '| correct_entities', '| wrong_entities',
             '| omitted_entities', '| redundant_entities', '| score']

    res_df = pd.DataFrame(r, index=index).T
    pd.set_option('precision', 4)
    pd.set_option('display.width', None)
    pd.set_option('display.max_columns', None)
    pd.set_option("colheader_justify", "center")
    print(res_df)

    print(
        f"正样本集得分为：{pos_correct_entities} / "
        f"({pos_correct_entities}+{pos_wrong_entities}+{pos_omitted_entities}+"
        f"{pos_redundant_entities}) = {round(pos_metric, 4)}，负样本集得分为：{neg_correct_dt} / ({neg_correct_dt} + "
        f"{neg_redundant_entities})={round(neg_metric, 4)}，",
        f"总体得分为： ({pos_correct_entities} + {neg_correct_dt}) / "
        f"({all_pos_entities}+{neg_correct_dt + neg_redundant_entities})={round(sum_metric_micro, 4)}",
        f"准确率：{accuracy}",
    )
    print('\n')
    if self_metric:
        more_not_error_pos = (pos_correct_entities + pos_redundant_entities) / (
                pos_correct_entities + pos_wrong_entities + pos_omitted_entities + pos_redundant_entities)
        f"自定义-正样本集得分为：{pos_correct_entities + pos_redundant_entities} /" \
        f" ({pos_correct_entities}+{pos_wrong_entities}+{pos_omitted_entities}+"
        f"{pos_redundant_entities}) = {round(more_not_error_pos, 4)}，负样本集得分为：{round(1, 4)}，"
        print('\n')
    return res_df  # type:pd.DataFrame



# encoding: utf-8
"""
@author: zyl
@file: sundat_match.py
@time: 2021/11/29 9:33
@desc:
"""


def sunday_match(target, pattern):
    """

    Args:
        target:
        pattern:

    Returns:

    """
    len_target = len(target)
    len_pattern = len(pattern)

    if len_pattern > len_target:
        return list()

    index = 0
    starts = []
    while index < len_target:
        if pattern == target[index:index + len_pattern]:
            starts.append(index)
            index += 1
        else:
            if (index + len(pattern)) >= len_target:
                return starts
            else:
                if target[index + len(pattern)] not in pattern:
                    index += (len_pattern + 1)
                else:
                    index += 1
    return starts


if __name__ == '__main__':
    t = "this is an apple , apple app app is app not app"
    t = t.split()
    p = "app not"
    p = p.split()
    print(t)
    print(p)
    print(sunday_match(target=t, pattern=p))



# encoding: utf-8
"""
@author: zyl
@file: t5.py
@time: 2021/11/29 14:46
@desc:
"""
import time

import pandas as pd
import wandb
from loguru import logger
from simpletransformers.t5 import T5Model

from zyl_utils.model_utils.metrics import ner_metric


class NerT5:
    """
    ner model for train and eval---t5--simple-trainsformers
    """

    def __init__(self):
        self.start_time = '...'
        self.end_time = '...'
        self.describe = " use simple-transformers--t5-model"

        self.wandb_proj = 'mt5'
        self.save_dir = './'  # save output_file
        self.model_version = 'v0.0.0.0'  # to save model or best model
        # like a,b,c,d : a 原始数据批次，b模型方法批次，c进行模型的处理的数据批次，d：迭代调参批次

        self.model_type = 't5'
        self.pretrained_model = 't5-base'  # 预训练模型位置 model_name

        self.use_cuda = True
        self.cuda_device = 0

        self.model_args = self.my_config()

    def my_config(self):
        return {
            'train_batch_size': 8,

            # multiprocess
            'use_multiprocessing': False,
            'use_multiprocessing_for_evaluation': False,

            # base config
            'reprocess_input_data': True,
            'use_cached_eval_features': False,
            'fp16': False,
            'manual_seed': 234,
            'gradient_accumulation_steps': 1,  # ::increase batch size,Use time for memory,

            # save
            'no_save': False,
            'save_eval_checkpoints': False,
            'save_model_every_epoch': False,
            'save_optimizer_and_scheduler': True,
            'save_steps': -1,

            # eval
            'evaluate_during_training': True,
            'evaluate_during_training_verbose': True,

            # normal
            'no_cache': False,
            'use_early_stopping': False,
            'encoding': None,
            'do_lower_case': False,
            'dynamic_quantize': False,
            'quantized_model': False,
            'silent': False,

            # save
            'overwrite_output_dir': True,
            'output_dir': self.save_dir + 'outputs/' + self.model_version + '/',
            'cache_dir': self.save_dir + 'cache/' + self.model_version + '/',
            'best_model_dir': self.save_dir + 'best_model/' + self.model_version + '/',
            'tensorboard_dir': self.save_dir + 'runs/' + self.model_version + '/' + time.strftime("%Y%m%d_%H%M%S",
                                                                                                  time.localtime()) + '/',

            # t5 args
            'use_multiprocessed_decoding': False,
            'num_beams': 1,
            'length_penalty': 2.0,
            'max_length': 20,
            'num_return_sequences': 1,
            'preprocess_inputs': True,
            'repetition_penalty': 1.0,
            'special_tokens_list': [],
            'top_k': None,
            'top_p': None,
        }

    @staticmethod
    def _deal_with_df(df):
        df = df[['prefix', 'input_text', 'target_text']]
        df = df.astype('str')
        return df

    def train(self, train_data: pd.DataFrame, eval_data: pd.DataFrame):
        # deal with dt
        train_data = NerT5._deal_with_df(train_data)
        eval_data = NerT5._deal_with_df(eval_data)
        train_size = train_data.shape[0]
        eval_size = eval_data.shape[0]

        all_steps = train_size / self.model_args.get('train_batch_size')
        self.model_args.update(
            {
                'train_size': train_size,
                'eval_size': eval_size,
                'logging_steps': int(max(all_steps / 10 / self.model_args.get('gradient_accumulation_steps'), 1)),
                'evaluate_during_training_steps': int(
                    max(all_steps / 10 / self.model_args.get('gradient_accumulation_steps'), 1)),
                'wandb_project': self.wandb_proj,
                'wandb_kwargs': {
                    'name': self.model_version + time.strftime("_%m%d_%H:%M:%S", time.localtime()),
                    'tags': [self.model_version, 'train']
                }
            }
        )

        model = T5Model(model_type=self.model_type, model_name=self.pretrained_model,
                        use_cuda=self.use_cuda, cuda_device=self.cuda_device, args=self.model_args)

        # train
        try:
            start_time = time.time()
            logger.info(f'start training: model_version---{self.model_version},train_size---{train_size}')
            model.train_model(train_data=train_data, eval_data=eval_data)
            logger.info('training finished!!!')
            end_time = time.time()
            logger.info(f'train time: {round(end_time - start_time, 4)} s')
        except Exception as error:
            logger.error(f'train failed!!! ERROR:{error}')
        finally:
            wandb.finish()
            # ModelUtils.remove_some_model_files(model.args)

    def eval(self, eval_data: pd.DataFrame, check_in_input_text: bool = False, delimiter='|',
             tokenizer=None, use_sliding_window=False, sliding_window=512, stride=0.8,
             pos_neg_ratio=None, use_multi_gpus=None, self_metric=False):
        # deal_with_dt
        eval_data = NerT5._deal_with_df(eval_data)
        eval_size = eval_data.shape[0]

        # wand_b
        wandb.init(project=self.wandb_proj, config=self.model_args,
                   name=self.model_version + time.strftime("_%m%d_%H:%M:%S", time.localtime()),
                   tags=[self.model_version, 'eval'])

        try:
            start_time = time.time()
            logger.info(f'start eval: model_version---{self.model_version},eval size---{eval_size}')

            model = T5Model(model_type=self.model_type, model_name=self.model_args.get('best_model_dir'),
                            use_cuda=self.use_cuda, cuda_device=self.cuda_device, args=self.model_args)

            eval_res = NerT5._eval_entity_recognition(model, eval_data=eval_data, delimiter=delimiter,
                                                      check_in_input_text=check_in_input_text,
                                                      tokenizer=tokenizer, use_sliding_window=use_sliding_window,
                                                      sliding_window=sliding_window, stride=stride,
                                                      pos_neg_ratio=pos_neg_ratio, use_multi_gpus=use_multi_gpus,
                                                      self_metric=self_metric
                                                      )

            wandb_log = {"eval_size": eval_size}
            for k, v in eval_res.items():
                wandb_log.update({k: v.iloc[2, -1]})
            wandb.log(wandb_log)

            logger.info('eval finished!!!')
            end_time = time.time()
            need_time = round((end_time - start_time) / eval_size, 5)
            eval_time = round(need_time * eval_size, 4)
            print(f'eval results: {eval_res}')
            logger.info(f'eval time: {need_time} s * {eval_size} = {eval_time} s')
        except Exception as error:
            logger.error(f'eval failed!!! ERROR:{error}')
        finally:
            wandb.finish()

    @staticmethod
    def _eval_entity_recognition(model, eval_data: pd.DataFrame, check_in_input_text: bool, delimiter='|',
                                 tokenizer=None, use_sliding_window=False, sliding_window=512, stride=0.8,
                                 pos_neg_ratio=None, use_multi_gpus=None, self_metric=False):
        """eval entity recognition in mt5 model, version-v2 , reference: https://docs.qq.com/doc/DYXRYQU1YbkVvT3V2

        Args:
            model: a mt5 model
            eval_data: a pd.Dataframe , must have columns ['prefix','input_text','target_text']
            check_in_input_text: if the entities are in input_texts
            delimiter: the delimiter in target_text to split different entities
            use_sliding_window: if truncate the input text when predict
            sliding_window: truncating_size
            stride: overlapping_size
            use_multi_gpus:use_multi_gpus
            pos_neg_ratio : the ratio of positive and negative sample importance
            self_metric:self_metric
            tokenizer: tokenizer to split sentence

        Returns:
            show report and res, {prefix:res_df},type:dict
        """
        eval_data = eval_data[['prefix', 'input_text', 'target_text']]
        eval_data = eval_data.astype('str')

        prefixes = eval_data['prefix'].to_list()
        input_texts = eval_data['input_text'].tolist()
        target_texts = eval_data['target_text'].tolist()
        revised_target_texts = NerT5._revise_target_texts(target_texts=target_texts,
                                                          input_texts=input_texts, delimiter=delimiter,
                                                          check_in_input_text=check_in_input_text)

        pred_target_texts = NerT5.predict_entity_recognition(model, prefixes, input_texts, tokenizer=tokenizer,
                                                             use_sliding_window=use_sliding_window,
                                                             sliding_window=sliding_window, stride=stride,
                                                             delimiter=delimiter, use_multi_gpus=use_multi_gpus)

        revised_pred_target_texts = NerT5._revise_target_texts(target_texts=pred_target_texts,
                                                               input_texts=input_texts, delimiter=delimiter,
                                                               check_in_input_text=check_in_input_text)

        eval_data['true_target_text'] = revised_target_texts
        eval_data['pred_target_text'] = revised_pred_target_texts

        eval_res = {}
        for prefix in set(prefixes):
            prefix_df = eval_data[eval_data['prefix'] == prefix]
            y_true = prefix_df['true_target_text'].tolist()
            y_pred = prefix_df['pred_target_text'].tolist()
            print(f'{prefix} report:')
            res_df = ner_metric.entity_recognition_v2(y_true, y_pred, pos_neg_ratio=pos_neg_ratio,
                                                      self_metric=self_metric)
            eval_res[prefix] = res_df

        print(f'sum report:')
        res_df = ner_metric.entity_recognition_v2(revised_target_texts, revised_pred_target_texts,
                                                  pos_neg_ratio=pos_neg_ratio, self_metric=self_metric)
        eval_res['sum'] = res_df

        return eval_res  # {prefix:res_df},type:dict

    @staticmethod
    def predict_entity_recognition(model, prefixes: list, input_texts: list, use_sliding_window=False,
                                   sliding_window=512, stride=0.8, tokenizer=None,
                                   delimiter='|', use_multi_gpus=None) -> list:
        """predict entity recognition in mt5 model,
        Args:
            model: a mt5 model
            prefixes: prefixes
            input_texts: input_texts
            use_sliding_window: if use_sliding_window
            sliding_window: sliding_window,the max token length for the model input(max_sequence_length)
            tokenizer: tokenizer
            stride: stride,(1-stride)*sliding_window for overlapping

            delimiter: the delimiter in target_text to split different entities,default: '|'
            use_multi_gpus: use_multi_gpus

        Returns:
            pred_target_texts:list,every element in pred_target_texts corresponds a prefix and an input_text
        """
        if len(input_texts) == 1:
            use_multi_gpus = None
        assert len(prefixes) == len(input_texts)

        if use_sliding_window:
            t_ids, t_prefixes, t_input_texts = NerT5._split_texts_with_sliding_window(input_texts, prefixes,
                                                                                      tokenizer=tokenizer,
                                                                                      sliding_window=sliding_window,
                                                                                      stride=stride)

            to_predict_texts = [i + ': ' + j for i, j in zip(t_prefixes, t_input_texts)]

            if not use_multi_gpus:
                pred_target_texts = model.predict(to_predict_texts)
            else:
                pred_target_texts = model.predict_gpu(to_predict_texts, gpus=use_multi_gpus)

            pred_target_texts = NerT5._combine_pred_target_texts_by_ids(pred_target_texts, t_ids, delimiter)
        else:
            to_predict_texts = [i + ': ' + j for i, j in zip(prefixes, input_texts)]
            if not use_multi_gpus:
                pred_target_texts = model.predict(to_predict_texts)
            else:
                pred_target_texts = model.predict_gpu(to_predict_texts, gpus=use_multi_gpus)
        assert len(pred_target_texts) == len(input_texts)
        return pred_target_texts  # type:list[str]

    @staticmethod
    def _split_text_with_sliding_window(text: str, sliding_window=128, tokenizer=None, stride=0.8) -> list:
        """ any sequence exceeding the max_seq_length will be split into several windows (sub-sequences),
        each of length max_seq_length. The windows will typically overlap each other to a certain degree to
        minimize any information loss that may be caused by hard cutoffs.


        Args:
            text: a str text
            sliding_window: truncating_size:sliding window, max_seq_length
            tokenizer: tokenizer
            stride: The amount of overlap between the windows,The stride can be specified in terms of either a fraction
             of the max_seq_length, or as an absolute number of tokens.

        Returns:
            truncated_input_text: the list of truncated_input_text
        """

        if not isinstance(text, str):
            text = str(text)

        if not tokenizer:
            try:
                from transformers.models.t5 import T5Tokenizer
                tokenizer = T5Tokenizer.from_pretrained("mt5-base")
            except Exception:
                print('no tokenizer....')

        tokens = tokenizer.tokenize(text)

        if len(tokens) <= sliding_window:
            return [text]
        else:
            split_text = []
            if stride < 1:
                step_size = int(sliding_window * stride)
            else:
                step_size = int(stride)
            steps = int(len(tokens) / step_size)
            for i in range(0, steps + 1):
                text_i_tokens = tokens[i * step_size:i * step_size + sliding_window]
                if text_i_tokens:
                    text_i = ''.join(text_i_tokens).replace('▁', ' ').strip()
                    split_text.append(text_i)

            if (len(split_text) > 1) and (
                    len(tokenizer.tokenize(split_text[-1])) < (sliding_window - step_size)):
                split_text = split_text[0:-1]
            return split_text

    @staticmethod
    def _split_texts_with_sliding_window(input_texts: list, prefixes: list, tokenizer=None,
                                         sliding_window=512, stride=0.8):
        """ for every input_text in input_texts, split it and record the split_ids for combining

        Args:
            input_texts: the list of many input_text
            prefixes: the prefix list of the input_texts list
            sliding_window: sliding_window,the max token length for the model input(max_sequence_length)
            tokenizer: tokenizer
            stride: stride,(1-stride)*sliding_window for overlapping

        Returns:
            split_ids, split_prefixes, split_input_texts
        """
        assert len(input_texts) == len(prefixes)  # every input_text corresponds a prefix
        input_texts_ids = range(len(input_texts))

        split_ids = []
        split_prefixes = []
        split_input_texts = []

        if not tokenizer:
            try:
                from transformers.models.t5 import T5Tokenizer
                tokenizer = T5Tokenizer.from_pretrained("mt5-base")
            except Exception:
                print('no tokenizer....')

        for i_t_d, p, i_t in zip(input_texts_ids, prefixes, input_texts):
            split_input_text = NerT5._split_text_with_sliding_window(i_t, sliding_window, tokenizer, stride)
            for t_i_t in split_input_text:
                split_ids.append(i_t_d)
                split_input_texts.append(t_i_t)
                split_prefixes.append(p)
        return split_ids, split_prefixes, split_input_texts  # type:tuple[list[int],list[str],list[str]]

    @staticmethod
    def _combine_pred_target_texts_by_ids(pred_target_texts, split_ids, delimiter: str = '|') -> list:
        """combine truncated_predicted_target_texts split_ids

        Args:
            pred_target_texts: the result of predicting the truncated input_texts
            split_ids: get the truncated_ids when truncating input_texts
            delimiter: the delimiter in target_text to split different entities

        Returns:
            pred_target_texts: predicted target_texts
        """
        ids_target_text_dict = dict()
        for i, j in zip(split_ids, pred_target_texts):
            if not ids_target_text_dict.get(i):
                ids_target_text_dict[i] = delimiter + j + delimiter
            else:
                ids_target_text_dict[i] = ids_target_text_dict[i] + j + delimiter

        pred_target_texts = [ids_target_text_dict[k] for k in sorted(ids_target_text_dict.keys())]
        return pred_target_texts  # type:list

    @staticmethod
    def _revise_target_texts(target_texts: list, input_texts: list, check_in_input_text: bool = False, delimiter='|'):
        """revise the target texts,

        Args:
            target_texts: the list of the target_texts
            input_texts:  the list of the input_texts
            check_in_input_text: if check the entities in input_text
            delimiter: the delimiter in target_text to split different entities

        Returns:
            revised_target_texts = list[set]
        """
        revised_target_texts = [NerT5._revise_target_text(t_t, return_format='set', delimiter=delimiter) for
                                t_t in target_texts]  # type:list[set,...]
        if check_in_input_text:
            revised_target_texts = NerT5._keep_entities_in_input_text(input_texts, revised_target_texts)
        return revised_target_texts  # type:list[set]

    @staticmethod
    def _revise_target_text(target_text: str, delimiter: str = '|', return_format='set'):
        """ revise the target text

        Args:
            target_text: str, target_text
            return_format: 'set' means:'every entity is an element in a set', 'str' means: different entities are split
                            by the delimiter
            delimiter: the delimiter in target_text to split different entities

        Returns:
            revised_target_text : set or list
        """
        assert isinstance(target_text, str)
        target_text = target_text.split(delimiter)
        target_text = set([' '.join(e.strip().split()) for e in target_text])
        if '' in target_text:
            target_text.remove('')
        if return_format == 'set':
            revised_target_text = target_text
        elif return_format == 'list':
            revised_target_text = list(target_text)
        else:  # return_format == 'str'
            revised_target_text = '|'
            if target_text != set():
                for entity in list(target_text):
                    revised_target_text += (str(entity) + '|')
        return revised_target_text

    @staticmethod
    def _keep_entities_in_input_text(input_texts: list, target_texts: list):
        """for each sample, for every entity ,keep the entities that are in the input text,and remove other entities

        Args:
            input_texts: the list of many input_text,and every input text is a string
            target_texts: the list of many target_text,and evert target text is a set

        Returns:
            revise_target_texts: list[str]
        """
        revised_target_texts = []
        for input_text, target_text in zip(input_texts, target_texts):
            if target_text != set():
                elements = list(target_text)
                for e in elements:
                    if str(e) not in input_text:
                        target_text.remove(e)  # type:set
            revised_target_texts.append(target_text)
        return revised_target_texts  # type:list[set]


if __name__ == '__main__':
    class Trainer(NerT5):
        def __init__(self):
            super(Trainer, self).__init__()
            self.wandb_proj = 'panel_entity_recognition'

        def train_sample(self):
            train_file = './test.xlsx'
            eval_file = './test.xlsx'
            train_df = pd.read_excel(train_file)  # type:pd.DataFrame
            eval_df = pd.read_excel(eval_file)  # type:pd.DataFrame
            print(len(train_df))
            print(len(eval_df))
            self.save_dir = './'
            self.model_version = 'v0.0.0.0'
            self.model_type = 'mt5'
            self.pretrained_model = 'google/mt5-base'  # 预训练模型位置 model_name
            self.use_cuda = True
            self.cuda_device = 0

            self.model_args = self.my_config()
            self.model_args.update(
                {
                    'train_file': train_file,
                    'eval_file': eval_file,
                    'num_train_epochs': 3,
                    'learning_rate': 3e-4,
                    'train_batch_size': 24,  # 28
                    'gradient_accumulation_steps': 16,
                    'eval_batch_size': 16,
                    'max_seq_length': 512,
                }
            )
            self.train(train_df, eval_df)


    class Evaluator(NerT5):
        def __init__(self):
            super(Evaluator, self).__init__()
            self.wandb_proj = 'panel_entity_recognition'

        def eval_sample(self):
            eval_file = './test.xlsx'
            eval_data = pd.read_excel(eval_file)

            self.save_dir = './'
            self.model_version = 'erv0.0.0.0'
            self.model_type = 'mt5'
            self.use_cuda = True
            self.cuda_device = 1

            self.model_args = self.my_config()
            self.model_args.update(
                {
                    'eval_file': eval_file,
                    'eval_batch_size': 16,
                    'max_seq_length': 512,
                }
            )
            self.eval(eval_data, check_in_input_text=False, delimiter='|',
                      tokenizer=None, use_sliding_window=False)



# encoding: utf-8
'''
@author: zyl
@file: api_utils.py
@time: 2021/11/8 17:48
@desc:
'''
from fastapi import FastAPI, Request
from fastapi.responses import HTMLResponse
from fastapi.staticfiles import StaticFiles
from fastapi.templating import Jinja2Templates

app = FastAPI()

app.mount("/static", StaticFiles(directory="static"), name="static")


templates = Jinja2Templates(directory="templates")


@app.get("/items/{id}", response_class=HTMLResponse)
async def read_item(request: Request, id: str):
    return templates.TemplateResponse("demo.html", {"request": request, "id": id})

if __name__ == '__main__':
    import uvicorn
    uvicorn.run(app, host="0.0.0.0", port=3243)




# encoding: utf-8
'''
@author: zyl
@file: utils.py
@time: 2021/11/11 9:35
@desc:
'''

import pandas as pd


def use_cmd_argument():
    import argparse

    parser = argparse.ArgumentParser(description='set some parameters')

    parser.add_argument('--', type=str, help='传入的数字',default='a')
    parser.add_argument('--', type=str, help='传入的数字', default='a')
    parser.add_argument('--', type=str, help='传入的数字', default='a')
    parser.add_argument('--', type=str, help='传入的数字', default='a')
    parser.add_argument('--', type=str, help='传入的数字', default='a')
    parser.add_argument('--', type=str, help='传入的数字', default='a')
    parser.add_argument('--', type=str, help='传入的数字', default='a')
    parser.add_argument('--', type=str, help='传入的数字', default='a')
    parser.add_argument('--', type=str, help='传入的数字', default='a')
    parser.add_argument('--', type=str, help='传入的数字', default='a')
    parser.add_argument('--', type=str, help='传入的数字', default='a')
    parser.add_argument('--', type=str, help='传入的数字', default='a')
    parser.add_argument('--', type=str, help='传入的数字', default='a')
    parser.add_argument('--', type=str, help='传入的数字', default='a')
    parser.add_argument('--', type=str, help='传入的数字', default='a')
    parser.add_argument('--', type=str, help='传入的数字', default='a')
    parser.add_argument('--', type=str, help='传入的数字', default='a')


    args = parser.parse_args()

    # 获得integers参数
    print(args.integers)
    return args



    parser.add_argument('integers', metavar='N', type=int, nargs='+',
                        help='an integer for the accumulator')
    parser.add_argument('--sum', dest='accumulate', action='store_const',
                        const=sum, default=max,
                        help='sum the integers (default: find the max)')

    args = parser.parse_args()
    print(args.accumulate(args.integers))




class Utils:
    def __init__(self):
        pass

    @staticmethod
    def to_excel(dataframe, excel_path, sheet_name='default'):
        try:
            from openpyxl import load_workbook
            book = load_workbook(excel_path)
            writer = pd.ExcelWriter(excel_path, engine='openpyxl')
            writer.book = book
        except:
            writer = pd.ExcelWriter(excel_path, engine='openpyxl')

        dataframe.to_excel(writer, sheet_name=sheet_name, index=False)
        writer.save()



# encoding: utf-8
'''
@author: zyl
@file: setup.py
@time: 2021/11/10 18:06
@desc:
'''
from setuptools import setup, find_packages

setup(
    name='zyl_utils',
    version='0.0.8',
    description=(
        '自用utils'
    ),
    long_description=open('README.md').read(),
    long_description_content_type = 'text/markdown',
    author='张玉良',
    author_email='1137379695@qq.com',
    maintainer='张玉良',
    maintainer_email='1137379695@qq.com',
    packages=find_packages(),
    platforms=["all"],
    url='https://github.com/ZYuliang/zyl-utils',
    install_requires=[
        "pandas",
        "transformers",
        "torch",
        "wandb",
        "loguru",
    ],
    classifiers=[
        'Programming Language :: Python :: 3',
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
)



# encoding: utf-8
"""
@author: zyl
@file: ner_utils.py
@time: 2021/9/14 14:03
@desc: ner utils for simple-transformer mt5 model, eval and predict
"""
import pandas as pd


class NERUtils:
    # ner utils for mt5 model
    def __init__(self):
        # eval_entity_recognition  ------评估
        #       revise_target_texts。
        #               revise_target_text
        #               keep_entities_in_input_text
        #       predict_entity_recognition-------预测
        #               split_texts_with_sliding_window
        #               model.predict_gpu
        #               combine_pred_target_texts_by_ids
        #       revise_target_texts
        #               revise_target_text
        #               keep_entities_in_input_text
        #       entity_recognition_v2-----标准
        pass

    @staticmethod
    def eval_entity_recognition(model, eval_df: pd.DataFrame, check_in_input_text: bool, delimiter='|', tokenizer=None,
                                use_sliding_window=False, sliding_window=512, stride=0.8, pos_neg_ratio=None,
                                use_multi_gpus=None, self_metric=False):
        """eval entity recognition in mt5 model, version-v2 , reference: https://docs.qq.com/doc/DYXRYQU1YbkVvT3V2

        Args:
            model: a mt5 model
            eval_df: a pd.Dataframe , must have columns ['prefix','input_text','target_text']
            check_in_input_text: if the entities are in input_texts
            delimiter: the delimiter in target_text to split different entities
            use_sliding_window: if truncate the input text when predict
            sliding_window: truncating_size
            stride: overlapping_size
            use_multi_gpus:use_multi_gpus
            pos_neg_ratio : the ratio of positive and negative sample importance
            self_metric:self_metric
            tokenizer: tokenizer to split sentence

        Returns:
            show report and res, {prefix:res_df},type:dict
        """

        prefixes = eval_df['prefix'].to_list()
        input_texts = eval_df['input_text'].tolist()
        target_texts = eval_df['target_text'].tolist()
        revised_target_texts = NERUtils.revise_target_texts(target_texts=target_texts,
                                                            input_texts=input_texts, delimiter=delimiter,
                                                            check_in_input_text=check_in_input_text)

        pred_target_texts = NERUtils.predict_entity_recognition(model, prefixes, input_texts, tokenizer=tokenizer,
                                                                use_sliding_window=use_sliding_window,
                                                                sliding_window=sliding_window, stride=stride,
                                                                delimiter=delimiter, use_multi_gpus=use_multi_gpus)

        revised_pred_target_texts = NERUtils.revise_target_texts(target_texts=pred_target_texts,
                                                                 input_texts=input_texts, delimiter=delimiter,
                                                                 check_in_input_text=check_in_input_text)

        eval_df['true_target_text'] = revised_target_texts
        eval_df['pred_target_text'] = revised_pred_target_texts

        eval_res = {}
        for prefix in set(prefixes):
            prefix_df = eval_df[eval_df['prefix'] == prefix]
            y_true = prefix_df['true_target_text'].tolist()
            y_pred = prefix_df['pred_target_text'].tolist()
            print(f'{prefix} report:')
            res_df = NERUtils.entity_recognition_v2(y_true, y_pred, pos_neg_ratio=pos_neg_ratio,
                                                    self_metric=self_metric)
            eval_res[prefix] = res_df

        print(f'sum report:')
        res_df = NERUtils.entity_recognition_v2(revised_target_texts, revised_pred_target_texts,
                                                pos_neg_ratio=pos_neg_ratio, self_metric=self_metric)
        eval_res['sum'] = res_df

        return eval_res  # {prefix:res_df},type:dict

    @staticmethod
    def predict_entity_recognition(model, prefixes: list, input_texts: list, use_sliding_window=False,
                                   sliding_window=512, stride=0.8, tokenizer=None,
                                   delimiter='|', use_multi_gpus=None) -> list:
        """predict entity recognition in mt5 model,
        Args:
            model: a mt5 model
            prefixes: prefixes
            input_texts: input_texts
            use_sliding_window: if use_sliding_window
            sliding_window: sliding_window,the max token length for the model input(max_sequence_length)
            tokenizer: tokenizer
            stride: stride,(1-stride)*sliding_window for overlapping

            delimiter: the delimiter in target_text to split different entities,default: '|'
            use_multi_gpus: use_multi_gpus

        Returns:
            pred_target_texts:list,every element in pred_target_texts corresponds a prefix and an input_text
        """
        if len(input_texts) == 1:
            use_multi_gpus = None
        assert len(prefixes) == len(input_texts)

        if use_sliding_window:
            t_ids, t_prefixes, t_input_texts = NERUtils.split_texts_with_sliding_window(input_texts, prefixes,
                                                                                        tokenizer=tokenizer,
                                                                                        sliding_window=sliding_window,
                                                                                        stride=stride)

            to_predict_texts = [i + ': ' + j for i, j in zip(t_prefixes, t_input_texts)]

            if not use_multi_gpus:
                pred_target_texts = model.predict(to_predict_texts)
            else:
                pred_target_texts = model.predict_gpu(to_predict_texts, gpus=use_multi_gpus)

            pred_target_texts = NERUtils.combine_pred_target_texts_by_ids(pred_target_texts, t_ids, delimiter)
        else:
            to_predict_texts = [i + ': ' + j for i, j in zip(prefixes, input_texts)]
            if not use_multi_gpus:
                pred_target_texts = model.predict(to_predict_texts)
            else:
                pred_target_texts = model.predict_gpu(to_predict_texts, gpus=use_multi_gpus)
        assert len(pred_target_texts) == len(input_texts)
        return pred_target_texts  # type:list[str]

    @staticmethod
    def split_text_with_sliding_window(text: str, sliding_window=128, tokenizer=None, stride=0.8) -> list:
        """ any sequence exceeding the max_seq_length will be split into several windows (sub-sequences),
        each of length max_seq_length. The windows will typically overlap each other to a certain degree to
        minimize any information loss that may be caused by hard cutoffs.


        Args:
            text: a str text
            sliding_window: truncating_size:sliding window, max_seq_length
            tokenizer: tokenizer
            stride: The amount of overlap between the windows,The stride can be specified in terms of either a fraction
             of the max_seq_length, or as an absolute number of tokens.

        Returns:
            truncated_input_text: the list of truncated_input_text
        """

        if not isinstance(text, str):
            text = str(text)

        if not tokenizer:
            try:
                from transformers.models.t5 import T5Tokenizer
                tokenizer = T5Tokenizer.from_pretrained("mt5-base")
            except Exception:
                print('no tokenizer....')

        tokens = tokenizer.tokenize(text)

        if len(tokens) <= sliding_window:
            return [text]
        else:
            split_text = []
            if stride < 1:
                step_size = int(sliding_window * stride)
            else:
                step_size = int(stride)
            steps = int(len(tokens) / step_size)
            for i in range(0, steps + 1):
                text_i_tokens = tokens[i * step_size:i * step_size + sliding_window]
                if text_i_tokens:
                    text_i = ''.join(text_i_tokens).replace('▁', ' ').strip()
                    split_text.append(text_i)

            if (len(split_text) > 1) and (
                    len(tokenizer.tokenize(split_text[-1])) < (sliding_window - step_size)):
                split_text = split_text[0:-1]
            return split_text

    @staticmethod
    def split_texts_with_sliding_window(input_texts: list, prefixes: list, tokenizer=None,
                                        sliding_window=512, stride=0.8):
        """ for every input_text in input_texts, split it and record the split_ids for combining

        Args:
            input_texts: the list of many input_text
            prefixes: the prefix list of the input_texts list
            sliding_window: sliding_window,the max token length for the model input(max_sequence_length)
            tokenizer: tokenizer
            stride: stride,(1-stride)*sliding_window for overlapping

        Returns:
            split_ids, split_prefixes, split_input_texts
        """
        assert len(input_texts) == len(prefixes)  # every input_text corresponds a prefix
        input_texts_ids = range(len(input_texts))

        split_ids = []
        split_prefixes = []
        split_input_texts = []

        if not tokenizer:
            try:
                from transformers.models.t5 import T5Tokenizer
                tokenizer = T5Tokenizer.from_pretrained("mt5-base")
            except Exception:
                print('no tokenizer....')

        for i_t_d, p, i_t in zip(input_texts_ids, prefixes, input_texts):
            split_input_text = NERUtils.split_text_with_sliding_window(i_t, sliding_window, tokenizer, stride)
            for t_i_t in split_input_text:
                split_ids.append(i_t_d)
                split_input_texts.append(t_i_t)
                split_prefixes.append(p)
        return split_ids, split_prefixes, split_input_texts  # type:tuple[list[int],list[str],list[str]]

    @staticmethod
    def combine_pred_target_texts_by_ids(pred_target_texts, split_ids, delimiter: str = '|') -> list:
        """combine truncated_predicted_target_texts split_ids

        Args:
            pred_target_texts: the result of predicting the truncated input_texts
            split_ids: get the truncated_ids when truncating input_texts
            delimiter: the delimiter in target_text to split different entities

        Returns:
            pred_target_texts: predicted target_texts
        """
        ids_target_text_dict = dict()
        for i, j in zip(split_ids, pred_target_texts):
            if not ids_target_text_dict.get(i):
                ids_target_text_dict[i] = delimiter + j + delimiter
            else:
                ids_target_text_dict[i] = ids_target_text_dict[i] + j + delimiter

        pred_target_texts = [ids_target_text_dict[k] for k in sorted(ids_target_text_dict.keys())]
        return pred_target_texts  # type:list

    @staticmethod
    def revise_target_texts(target_texts: list, input_texts: list, check_in_input_text: bool = False, delimiter='|'):
        """revise the target texts,

        Args:
            target_texts: the list of the target_texts
            input_texts:  the list of the input_texts
            check_in_input_text: if check the entities in input_text
            delimiter: the delimiter in target_text to split different entities

        Returns:
            revised_target_texts = list[set]
        """
        revised_target_texts = [NERUtils.revise_target_text(t_t, return_format='set', delimiter=delimiter) for
                                t_t in target_texts]  # type:list[set,...]
        if check_in_input_text:
            revised_target_texts = NERUtils.keep_entities_in_input_text(input_texts, revised_target_texts)
        return revised_target_texts  # type:list[set]

    @staticmethod
    def revise_target_text(target_text: str, delimiter: str = '|', return_format='set'):
        """ revise the target text

        Args:
            target_text: str, target_text
            return_format: 'set' means:'every entity is an element in a set', 'str' means: different entities are split
                            by the delimiter
            delimiter: the delimiter in target_text to split different entities

        Returns:
            revised_target_text : set or list
        """
        assert isinstance(target_text, str)
        target_text = target_text.split(delimiter)
        target_text = set([' '.join(e.strip().split()) for e in target_text])
        if '' in target_text:
            target_text.remove('')
        if return_format == 'set':
            revised_target_text = target_text
        elif return_format == 'list':
            revised_target_text = list(target_text)
        else:  # return_format == 'str'
            revised_target_text = '|'
            if target_text != set():
                for entity in list(target_text):
                    revised_target_text += (str(entity) + '|')
        return revised_target_text

    @staticmethod
    def keep_entities_in_input_text(input_texts: list, target_texts: list):
        """for each sample, for every entity ,keep the entities that are in the input text,and remove other entities

        Args:
            input_texts: the list of many input_text,and every input text is a string
            target_texts: the list of many target_text,and evert target text is a set

        Returns:
            revise_target_texts: list[str]
        """
        revised_target_texts = []
        for input_text, target_text in zip(input_texts, target_texts):
            if target_text != set():
                elements = list(target_text)
                for e in elements:
                    if str(e) not in input_text:
                        target_text.remove(e)  # type:set
            revised_target_texts.append(target_text)
        return revised_target_texts  # type:list[set]

    @staticmethod
    def entity_recognition_v2(y_true: list, y_pred: list, pos_neg_ratio: str = None, self_metric=False):
        """the metric of entity_recognition, version-v2, reference: https://docs.qq.com/doc/DYXRYQU1YbkVvT3V2

        Args:
            y_true: list[set],the list of true target texts,each element is a set
            y_pred: list[set],the list of pred target texts,each element is a set
            pos_neg_ratio: the ratio of positive and negative sample importance, default: the ratio of positive and
                           negative sample sizes, you can set it,like"7:3"
            self_metric: self_metric

        Returns:
            show report and res
        """

        neg_data = 0
        neg_correct_dt = 0
        neg_wrong_dt = 0
        neg_redundant_entities = 0

        pos_data = 0
        pos_correct_dt = 0
        pos_wrong_dt = 0
        pos_correct_entities = 0
        pos_wrong_entities = 0
        pos_omitted_entities = 0
        pos_redundant_entities = 0

        for i, j in zip(y_true, y_pred):
            if i == set():
                neg_data += 1
                if j == set():
                    neg_correct_dt += 1
                else:
                    neg_wrong_dt += 1
                    neg_redundant_entities += len(j)
            else:
                pos_data += 1
                true_pred = len(i & j)
                pos_correct_entities += true_pred

                if i == j:
                    pos_correct_dt += 1
                elif len(i) >= len(j):
                    pos_wrong_dt += 1
                    pos_wrong_entities += (len(j) - true_pred)
                    pos_omitted_entities += (len(i) - len(j))
                else:
                    pos_wrong_dt += 1
                    pos_redundant_entities += (len(j) - len(i))
                    pos_wrong_entities += (len(i) - true_pred)

        all_pos_entities = pos_correct_entities + pos_wrong_entities + pos_omitted_entities + pos_redundant_entities
        if neg_data == 0:
            neg_metric = 0
        else:
            neg_metric = neg_correct_dt / (neg_correct_dt + neg_redundant_entities)
        if pos_data == 0:
            pos_metric = 0
        else:
            pos_metric = pos_correct_entities / all_pos_entities

        sum_metric_micro = (pos_correct_entities + neg_correct_dt) / (
                neg_correct_dt + neg_redundant_entities + all_pos_entities)
        # sum_metric_macro = neg_metric * 0.5 + pos_metric * 0.5

        if pos_neg_ratio:
            pos_all = float(pos_neg_ratio.split(':')[0])
            neg_all = float(pos_neg_ratio.split(':')[1])
            pos_ratio = pos_all / (pos_all + neg_all)
            neg_ratio = neg_all / (pos_all + neg_all)
        else:
            pos_ratio = pos_data / (pos_data + neg_data)
            neg_ratio = neg_data / (pos_data + neg_data)

        sum_metric_weighted = pos_ratio * pos_metric + neg_ratio * neg_metric

        # pos_precision = pos_correct_dt / (neg_correct_dt + pos_correct_dt)
        # recall = pos_correct_dt / pos_data
        tp = pos_correct_dt
        fn = pos_wrong_dt
        fp = neg_wrong_dt
        tn = neg_correct_dt

        accuracy = (tp + tn) / (tp + fn + fp + tn)
        # precision = tp / (tp + fp)
        # recall = tp / (tp + fn)
        # f1 = 2 / (1 / precision + 1 / recall)
        r = {
            'positive data': [str(pos_data), pos_correct_dt, pos_wrong_dt, pos_correct_entities,
                              pos_wrong_entities, pos_omitted_entities, pos_redundant_entities, pos_metric],
            'negative data': [neg_data, neg_correct_dt, neg_wrong_dt, '-', '-', '-', neg_redundant_entities,
                              neg_metric],

            'all data ': [str(pos_data + neg_data), neg_correct_dt + pos_correct_dt, neg_wrong_dt + pos_wrong_dt,
                          pos_correct_entities, pos_wrong_entities, pos_omitted_entities,
                          pos_redundant_entities + neg_redundant_entities,
                          sum_metric_micro],
            # 'precision': ['', '', '', '', '', '', '', precision],
            # 'recall': ['', '', '', '', '', '', '', recall],
            # 'f1 score': ['', '', '', '', '', '', '', (2 * precision * recall) / (precision + recall)],
            # 'accuracy score': ['', '', '', '', '', '', '', (neg_correct_dt + pos_correct_dt) / (pos_data + neg_data)],
            # 'micro score': ['', '', '', '', '', '', '', sum_metric_micro],
            # 'macro score': ['', '', '', '', '', '', '', sum_metric_macro],
            'weighted score': ['', '', '', '', '', '', '', sum_metric_weighted],
        }

        index = ['| data_num', '| correct_data', '| wrong_data', '| correct_entities', '| wrong_entities',
                 '| omitted_entities', '| redundant_entities', '| score']

        res_df = pd.DataFrame(r, index=index).T
        pd.set_option('precision', 4)
        pd.set_option('display.width', None)
        pd.set_option('display.max_columns', None)
        pd.set_option("colheader_justify", "center")
        print(res_df)

        print(
            f"正样本集得分为：{pos_correct_entities} / "
            f"({pos_correct_entities}+{pos_wrong_entities}+{pos_omitted_entities}+"
            f"{pos_redundant_entities}) = {round(pos_metric, 4)}，负样本集得分为：{neg_correct_dt} / ({neg_correct_dt} + "
            f"{neg_redundant_entities})={round(neg_metric, 4)}，",
            f"总体得分为： ({pos_correct_entities} + {neg_correct_dt}) / "
            f"({all_pos_entities}+{neg_correct_dt + neg_redundant_entities})={round(sum_metric_micro, 4)}",
            f"准确率：{accuracy}",
        )
        print('\n')
        if self_metric:
            more_not_error_pos = (pos_correct_entities + pos_redundant_entities) / (
                    pos_correct_entities + pos_wrong_entities + pos_omitted_entities + pos_redundant_entities)
            f"自定义-正样本集得分为：{pos_correct_entities + pos_redundant_entities} /" \
            f" ({pos_correct_entities}+{pos_wrong_entities}+{pos_omitted_entities}+"
            f"{pos_redundant_entities}) = {round(more_not_error_pos, 4)}，负样本集得分为：{round(1, 4)}，"
            print('\n')
        return res_df  # type:pd.DataFrame


if __name__ == '__main__':
    pass



# encoding: utf-8
'''
@author: zyl
@file: T5_model.py
@time: 2021/11/11 10:54
@desc:
'''
import copy
from concurrent.futures import ThreadPoolExecutor, as_completed

import torch
from simpletransformers.t5 import T5Model, DDPT5Model

from zyl_utils.data_utils.nlp_utils import DTUtils


class MyT5(T5Model):
    """
    add function: use-multi-gpu
    """

    def __init__(self, model_type, model_name, args=None, tokenizer=None, use_cuda=True, cuda_device=-1, **kwargs):
        super(MyT5, self).__init__(model_type=model_type, model_name=model_name, args=args,
                                   tokenizer=tokenizer, use_cuda=use_cuda, cuda_device=cuda_device, **kwargs)

    def get_funcs(self, gpus):
        self.funcs = []
        for i in gpus:
            if i != self.device.index:
                other_m = copy.deepcopy(self)
                other_m.device = torch.device(f"cuda:{i}")
                self.funcs.append(other_m.predict)
            else:
                self.funcs.append(self.predict)

    def predict_gpu(self, to_predict, gpus: list = None):
        # gpus can be like： ["1","2"]
        if len(to_predict) <= len(gpus):
            gpus = None
        if gpus and (len(gpus) == 1):
            gpus = None
        if not gpus:
            outputs = self.predict(to_predict=to_predict)
        else:
            if not self.funcs:
                self.get_funcs(gpus)
            print('Start processing data...')
            max_workers = len(gpus)
            sub_data_sets = DTUtils.split_data_evenly(to_predict, len(gpus))
            res = dict()
            with ThreadPoolExecutor(max_workers=max_workers) as executor:
                assert len(self.funcs) == len(sub_data_sets)
                futures = {executor.submit(self.funcs[n], dt): n for dt, n in
                           zip(sub_data_sets, list(range(len(sub_data_sets))))}

                for f in as_completed(futures):  # not block,iterator
                    f.dt_id = futures[f]
                    res.update({f.dt_id: f.result()})
            outputs = []
            for i in sorted(res.keys()):
                for j in res[i]:
                    outputs.append(j)
        return outputs


class MyDDPT5(DDPT5Model):
    """
    add function: use-multi-gpu
    """

    def __init__(self, model_type, model_name, args=None, tokenizer=None, use_cuda=True, cuda_device=-1, **kwargs):
        super(MyDDPT5, self).__init__(model_type=model_type, model_name=model_name, args=args,
                                      tokenizer=tokenizer, use_cuda=use_cuda, cuda_device=cuda_device, **kwargs)

    def get_funcs(self, gpus):
        self.funcs = []
        for i in gpus:
            if i != self.device.index:
                other_m = copy.deepcopy(self)
                other_m.device = torch.device(f"cuda:{i}")
                self.funcs.append(other_m.predict)
            else:
                self.funcs.append(self.predict)

    def predict_gpu(self, to_predict, gpus: list = None):
        # gpus can be like： ["1","2"]
        if len(to_predict) <= len(gpus):
            gpus = None
        if gpus and (len(gpus) == 1):
            gpus = None
        if not gpus:
            outputs = self.predict(to_predict=to_predict)
        else:
            if not self.funcs:
                self.get_funcs(gpus)
            print('Start processing data...')
            max_workers = len(gpus)
            sub_data_sets = DTUtils.split_data_evenly(to_predict, len(gpus))
            res = dict()
            with ThreadPoolExecutor(max_workers=max_workers) as executor:
                assert len(self.funcs) == len(sub_data_sets)
                futures = {executor.submit(self.funcs[n], dt): n for dt, n in
                           zip(sub_data_sets, list(range(len(sub_data_sets))))}

                for f in as_completed(futures):  # not block,iterator
                    f.dt_id = futures[f]
                    res.update({f.dt_id: f.result()})
            outputs = []
            for i in sorted(res.keys()):
                for j in res[i]:
                    outputs.append(j)
        return outputs
